from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Type, Union

import genv.utils


@dataclass
class Env:
    @dataclass
    class Config:
        name: Optional[str] = None
        gpu_memory: Optional[str] = None
        gpus: Optional[int] = None

        def load(self, path: Union[str, Path]) -> None:
            """Loads from disk"""

            def _load_field(basename: str, type: Type = str) -> None:
                file = Path(path).joinpath(basename)

                if file.is_file():
                    return type(file.read_text().strip())

            self.name = _load_field("name")
            self.gpu_memory = _load_field("gpu-memory")
            self.gpus = _load_field("gpus", int)

        def save(self, path: Union[str, Path]) -> None:
            """Saves to disk"""

            def _save_field(basename: str, value: Optional[Any]) -> None:
                file = Path(path).joinpath(basename)

                if value is not None:
                    file.write_text(str(value))
                else:
                    file.unlink(missing_ok=True)

            _save_field("name", self.name)
            _save_field("gpu-memory", self.gpu_memory)
            _save_field("gpus", self.gpus)

    eid: str
    uid: int
    creation: str
    username: Optional[str]
    config: Config
    pids: Iterable[int]
    kernel_ids: Iterable[str]

    @property
    def time_since(self) -> str:
        return genv.utils.time_since(self.creation)

    @property
    def active(self) -> bool:
        return len(self.pids) > 0 or len(self.kernel_ids) > 0

    def __hash__(self) -> int:
        return self.eid.__hash__()

    def cleanup(
        self,
        *,
        poll_pid: Callable[[int], bool] = None,
        poll_kernel: Callable[[str], bool] = None,
    ):
        """
        Cleans up in place.
        """
        if poll_pid is not None:
            self.pids = [pid for pid in self.pids if poll_pid(pid)]

        if poll_kernel is not None:
            self.kernel_ids = [
                kernel_id for kernel_id in self.kernel_ids if poll_kernel(kernel_id)
            ]

    def attach(
        self, *, pid: Optional[int] = None, kernel_id: Optional[str] = None
    ) -> None:
        """
        Attaches a process or a Jupyter kernel to an environment.
        """
        if pid is not None:
            self.pids.append(pid)

        if kernel_id is not None:
            self.kernel_ids.append(kernel_id)


@dataclass
class Envs:
    """
    A collection of environments.
    """

    envs: Iterable[Env]

    @property
    def eids(self) -> Iterable[str]:
        return [env.eid for env in self.envs]

    @property
    def usernames(self) -> Iterable[str]:
        return set(env.username for env in self.envs if env.username)

    def __iter__(self):
        return self.envs.__iter__()

    def __len__(self):
        return self.envs.__len__()

    def __getitem__(self, eid: str) -> Env:
        return next(env for env in self.envs if env.eid == eid)

    def __contains__(self, eid: str) -> bool:
        return any(env.eid == eid for env in self.envs)

    def activate(
        self,
        eid: str,
        uid: int,
        username: Optional[str] = None,
    ) -> None:
        """
        Activates a new environment.
        """
        self.envs.append(
            Env(
                eid=eid,
                uid=uid,
                creation=datetime.now().strftime(genv.utils.DATETIME_FMT),
                username=username,
                config=Env.Config(name=None, gpu_memory=None, gpus=None),
                pids=[],
                kernel_ids=[],
            )
        )

    def filter(
        self,
        deep: bool = True,
        *,
        eid: Optional[str] = None,
        eids: Optional[Iterable[str]] = None,
        username: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Returns a new filtered collection.

        :param deep: Perform deep filtering
        :param eid: Environment identifier to keep
        :param eids: Environment identifiers to keep
        :param username: Username to keep
        """
        if eids:
            eids = set(eids)

        if eid:
            if not eids:
                eids = set()

            eids.add(eid)

        envs = self.envs

        if eids is not None:
            envs = [env for env in envs if env.eid in eids]

        if username is not None:
            envs = [env for env in envs if env.username == username]

        if name is not None:
            envs = [env for env in envs if env.config.name == name]

        return Envs(envs)

    def cleanup(
        self,
        *,
        eid: Optional[str] = None,
        eids: Optional[Iterable[str]] = None,
        poll_pid: Callable[[int], bool] = None,
        poll_kernel: Callable[[str], bool] = None,
    ):
        """Cleans up the collection in place.

        Cleans only the specified environments when passing identifiers.
        """
        if eids:
            eids = set(eids)

        if eid:
            if not eids:
                eids = set()

            eids.add(eid)

        for env in self.envs:
            if eids is not None:
                if env.eid not in eids:
                    continue

            env.cleanup(poll_pid=poll_pid, poll_kernel=poll_kernel)

        self.envs = [env for env in self.envs if env.active]

    def find(
        self, *, pid: Optional[int] = None, kernel_id: Optional[str] = None
    ) -> Iterable[Env]:
        """
        Returns the environments of the given process or kernel.
        """

        def _pred(env: Env) -> bool:
            if (pid is not None) and (pid in env.pids):
                return True

            if (kernel_id is not None) and (kernel_id in env.kernel_ids):
                return True

            return False

        return [env for env in self.envs if _pred(env)]



import argparse
import sys
import time

import genv


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv monitor" arguments to a parser.
    """

    parser.add_argument(
        "--config-dir",
        default=genv.utils.get_temp_file_path("metrics"),
        help="Directory to create Prometheus and Grafana config files at (default: %(default)s)",
    )

    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=8000,
        help="Port for Prometheus exporter to listen on (default: %(default)s)",
    )

    parser.add_argument(
        "-i",
        "--interval",
        type=int,
        default=10,
        help="Interval in seconds between collections (default: %(default)s)",
    )


async def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv monitor" logic.
    """

    # NOTE(raz): we import depdendencies only here as they are optional
    try:
        from genv.metrics import Collection, publish_config_files, SPECS
        import prometheus_client
    except ModuleNotFoundError as e:
        if e.name != "prometheus_client":
            raise

        print(f"ERROR: Python package '{e.name}' is required", file=sys.stderr)
        exit(1)

    # https://github.com/prometheus/client_python#disabling-default-collector-metrics
    prometheus_client.REGISTRY.unregister(prometheus_client.GC_COLLECTOR)
    prometheus_client.REGISTRY.unregister(prometheus_client.PLATFORM_COLLECTOR)
    prometheus_client.REGISTRY.unregister(prometheus_client.PROCESS_COLLECTOR)

    prometheus_client.start_http_server(args.port)

    publish_config_files(args.config_dir, prometheus_exporter_port=args.port)

    collection = Collection(SPECS)

    while True:
        system = await genv.core.system()

        with genv.utils.global_lock():
            snapshot = await genv.core.snapshot()

        collection.cleanup(system, snapshot)
        collection.update(system, snapshot)

        time.sleep(args.interval)



from genv.utils import memory_to_bytes, bytes_to_memory

from genv.entities.enforce import Survey


def env_memory(*surveys: Survey) -> None:
    """
    Terminates processes from environments that exceed their memory capacity.
    """
    for survey in surveys:
        for env in survey.snapshot.envs:
            if env.config.gpu_memory is None:
                continue

            for device in survey.snapshot.devices.filter(eid=env.eid):
                processes = survey.snapshot.processes.filter(
                    eid=env.eid, index=device.index
                )

                used_bytes = sum(process.total_bytes for process in processes)

                over_bytes = used_bytes - memory_to_bytes(env.config.gpu_memory)

                if not over_bytes > 0:
                    continue

                print(
                    f"{f'[{survey.hostname}] ' if survey.hostname else ''}"
                    f"Environment {env.eid} is using {bytes_to_memory(used_bytes, 'm')} on device {device.index} which is {bytes_to_memory(over_bytes, 'm')} over its capacity of {env.config.gpu_memory}"
                )

                freed_bytes = 0

                for process in processes:
                    survey.terminate(process.pid)

                    freed_bytes += process.total_bytes

                    if freed_bytes >= over_bytes:
                        break



import argparse
import subprocess
import sys
from typing import NoReturn

import genv.sdk


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv lock" arguments to a parser.
    """

    parser.add_argument(
        "args",
        nargs=argparse.REMAINDER,
        metavar="COMMAND",
        help="Command to run under the device lock",
    )


def run(args: argparse.Namespace) -> NoReturn:
    """
    Runs the "genv lock" logic.
    """

    with genv.sdk.devices.lock():
        status = subprocess.run(args.args).returncode if len(args.args) else 0
        sys.exit(status)



import argparse

import genv.sdk


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv attach" arguments to a parser.
    """

    modes = parser.add_argument_group("modes")
    group = modes.add_mutually_exclusive_group()
    group.add_argument(
        "--config",
        action="store_true",
        default=True,
        help="Use device count from configuration (default)",
    )
    group.add_argument(
        "--count", type=int, help="Total number of devices to be attached"
    )
    group.add_argument(
        "--index", type=int, help="Attach to the device with the given index"
    )
    group.add_argument(
        "--refresh", action="store_true", help="Only refresh attachments"
    )

    options = parser.add_argument_group("options")
    options.add_argument(
        "-o",
        "--over-subscribe",
        dest="allow_over_subscription",
        action="store_true",
        help="Use unavailable devices if needed",
    )


def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv attach" logic.
    """

    if not genv.sdk.active():
        raise RuntimeError("Not running in an active environment")

    if not args.refresh:
        genv.sdk.attach(
            index=args.index,
            gpus=args.count,
            allow_over_subscription=args.allow_over_subscription,
        )



import asyncio
from dataclasses import dataclass
import sys
from typing import Any, Callable, Iterable, Optional, Tuple

from genv.utils.runners import SSH as Runner

from .utils import reprint


@dataclass
class Host:
    """
    Host configuration.

    :param hostname: Hostname or IP address
    :param username: SSH login name
    :param timeout: SSH connection timeout
    """

    hostname: str
    username: Optional[str]
    timeout: Optional[int]


@dataclass
class Config:
    """
    Execution configuration.

    :param hosts: Host configurations
    :param throw_on_error: Raise 'RuntimeError' if failing to connect to any host
    :param quiet: Ignore SSH errors
    """

    hosts: Iterable[Host]
    throw_on_error: bool
    quiet: bool


@dataclass
class Command:
    """
    Genv command specification.

    :param args: Genv command arguments
    :param sudo: Run command as root using sudo
    :param shell: Run command as regular shell command
    """

    args: Iterable[str]
    sudo: bool = False
    shell: bool = False

    @property
    def all_args(self) -> Iterable[str]:
        """Returns all command arguments"""

        return self.args if self.shell else ["genv"] + self.args


async def run(
    config: Config,
    command: Command,
    stdins: Optional[Iterable[str]] = None,
) -> Tuple[Iterable[Host], Iterable[str]]:
    """
    Runs a command on multiple hosts over SSH.
    Waits for the command to finish on all hosts.
    Raises 'RuntimeError' if failed to connect to any of the hosts and 'config.throw_on_error' is True.

    :param config: Execution configuration
    :param command: Command specification
    :param stdins: Input to send per host

    :return: Returns the hosts that succeeded and their standard outputs
    """

    runners = [Runner(host.hostname, host.username, host.timeout) for host in config.hosts]

    results = await asyncio.gather(
        *(
            runner.run(*command.all_args, stdin=stdin, sudo=command.sudo, check=False)
            for runner, stdin in zip(runners, stdins or [None for _ in runners])
        )
    )

    processes = [result.process for result in results]
    stdouts = [result.stdout for result in results]
    stderrs = [result.stderr for result in results]

    def filter(
        objs: Iterable[Any], pred: Callable[[asyncio.subprocess.Process], bool]
    ) -> Iterable[Any]:
        return [obj for process, obj in zip(processes, objs) if pred(process)]

    def succeeded(objs: Iterable[Any]) -> Iterable[Any]:
        return filter(objs, lambda process: process.returncode == 0)

    def failed(objs: Iterable[Any]) -> Iterable[Any]:
        return filter(objs, lambda process: process.returncode != 0)

    for host, stderr in failed(zip(config.hosts, stderrs)):
        message = f"Failed running SSH command on {host.hostname} ({stderr})"

        if config.throw_on_error:
            raise RuntimeError(message)
        elif not config.quiet:
            print(message, file=sys.stderr)

    hosts = succeeded(config.hosts)
    stdouts = succeeded(stdouts)
    stderrs = succeeded(stderrs)

    reprint([host.hostname for host in hosts], stderrs, file=sys.stderr)

    return hosts, stdouts



from . import utils
from .entities import (
    Device,
    Devices,
    Env,
    Envs,
    Process,
    Processes,
    Report,
    Snapshot,
    Survey,
    System,
)
from .serialization import JSONEncoder, JSONDecoder
from . import core
from . import enforce
from . import remote
from . import sdk
from . import cli



import os
from typing import Dict, Iterable

from .runners import Runner, Local

DEFAULT_NVIDIA_SMI_ENV_VARS = {"GENV_BYPASS": "1", **os.environ}
DEFAULT_LOCAL_RUNNER = Local(DEFAULT_NVIDIA_SMI_ENV_VARS)

NVIDIA_SMI = "nvidia-smi"
CSV_FORMAT_PARAM = "--format=csv,noheader,nounits"
NVIDIA_CSV_SPLITTER = ", "


async def device_uuids(runner: Runner = DEFAULT_LOCAL_RUNNER) -> Dict[str, int]:
    """
    Queries device UUIDs.

    :return: A mapping from device UUID to its index
    """
    stdout = (await runner.run(NVIDIA_SMI, "--query-gpu=uuid,index", CSV_FORMAT_PARAM, check=True)).stdout

    mapping = dict()

    for line in stdout.splitlines():
        uuid, index = line.split(NVIDIA_CSV_SPLITTER)
        mapping[uuid] = int(index)

    return mapping


async def compute_apps(runner: Runner = DEFAULT_LOCAL_RUNNER) -> Iterable[Dict]:
    """
    Queries the running compute apps.
    """
    output = (
        await runner.run(
            NVIDIA_SMI,
            "--query-compute-apps=gpu_uuid,pid,used_gpu_memory",
            CSV_FORMAT_PARAM,
            check=True)
    ).stdout

    apps = []

    for line in output.splitlines():
        gpu_uuid, pid, used_gpu_memory = line.split(NVIDIA_CSV_SPLITTER)

        apps.append(
            dict(
                gpu_uuid=gpu_uuid, pid=int(pid), used_gpu_memory=f"{used_gpu_memory}mi"
            )
        )

    return apps



import argparse

import genv.sdk


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv status" arguments to a parser.
    """

    pass


def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv status" logic.
    """

    if not genv.sdk.active():
        print("Environment is not active")
    else:
        print(f"Environment is active ({genv.sdk.eid()})")

        indices = genv.sdk.attached()

        if not indices:
            print("Detached from GPUs")
        else:
            print(f"Attached to GPUs at indices {','.join(map(str, indices))}")

        config = genv.sdk.configuration()

        print()
        print("Configuration")
        print(f"    Name: {config.name or 'N/A'}")
        print(f"    Device count: {config.gpus or 'N/A'}")
        print(f"    GPU memory capacity: {config.gpu_memory or 'N/A'}")
        print()



from contextlib import contextmanager
import os
from typing import Any, Iterable, Optional, Union


def get_temp_env_vars(
    *, ignore: Optional[Union[str, Iterable[str]]] = None
) -> Iterable[str]:
    """Returns the temporary variables"""

    if ignore is None:
        ignore = []
    elif isinstance(ignore, str):
        ignore = [ignore]

    return [
        name
        for name in os.environ.get("GENV_ENVS", "").split(":")
        if (len(name) and (name not in ignore))
    ]


def set_temp_env_var(name: str, value: Any) -> None:
    """Sets a temporary environment variable"""

    os.environ[name] = str(value)
    os.environ["GENV_ENVS"] = ":".join(get_temp_env_vars(ignore=name) + [name])


def unset_temp_env_var(name: str) -> None:
    """Unsets a temporary environment variable if set"""

    os.environ.pop(name, None)
    os.environ["GENV_ENVS"] = ":".join(get_temp_env_vars(ignore=name))


def unset_temp_env_vars() -> None:
    """Unsets all temporary environment variables"""

    for name in get_temp_env_vars():
        unset_temp_env_var(name)

    os.environ.pop("GENV_ENVS", None)


@contextmanager
def temp_env_vars() -> None:
    """Context manager for unsetting temporary environment variables"""

    try:
        yield
    finally:
        unset_temp_env_vars()



from typing import Any, Optional, Union

import genv.utils
from genv.entities import Env, Envs
import genv.serialization

from .utils import File


class State(File[Envs]):
    def __init__(self, cleanup: bool = True, reset: bool = False) -> None:
        super().__init__(
            genv.utils.get_temp_file_path("envs.json"),
            cleanup,
            reset,
        )

    def _create(self):
        return Envs([])

    def _convert(self, o: Union[Any, Envs]) -> Envs:
        def _get_field(obj, field, cls, *args):
            return (
                getattr(obj, field, *args)
                if isinstance(obj, cls)
                else obj.get(field, *args)
            )

        def _get_env_field(env, field, *args):
            return _get_field(env, field, Env, *args)

        def _get_config_field(env, field, *args):
            config = _get_env_field(env, "config")

            return _get_field(config, field, Env.Config, *args)

        # the following logic converts state files from versions <= 0.9.0.
        # note that the indicator is 'o.envs' and not 'o' itself because the structure of
        # the state file from these versions was similar to the snapshot structure (a single
        # key named "envs") so the JSON decoder parsed the object as a snapshot object.
        if isinstance(o.envs, dict):
            o = Envs(
                [
                    Env(
                        _get_env_field(env, "eid"),
                        _get_env_field(env, "uid"),
                        _get_env_field(env, "creation"),
                        _get_env_field(env, "username", None),
                        Env.Config(
                            _get_config_field(env, "name", None),
                            _get_config_field(env, "gpu_memory", None),
                            _get_config_field(env, "gpus", None),
                        ),
                        _get_env_field(env, "pids", []),
                        _get_env_field(env, "kernel_ids", []),
                    )
                    for env in o.envs.values()
                ]
            )

        return o

    def _clean(self, envs: Envs) -> None:
        envs.cleanup(
            poll_pid=genv.utils.poll.poll_pid,
            poll_kernel=genv.utils.poll.poll_jupyter_kernel,
        )


def snapshot() -> Envs:
    """
    Returns an environments snapshot.
    """
    return State().load()


def activate(
    eid: str,
    uid: int,
    username: Optional[str] = None,
    *,
    pid: Optional[int] = None,
    kernel_id: Optional[str] = None,
) -> None:
    """
    Activates an environment if does not exist and attaches a proces or a Jupyter kernel to it.
    """
    with State() as envs:
        if eid not in envs:
            envs.activate(
                eid=eid,
                uid=uid,
                username=username,
            )

        envs[eid].attach(pid=pid, kernel_id=kernel_id)


def configuration(eid: str) -> Optional[Env.Config]:
    """Returns the configuration of an environment if it exists"""

    envs = snapshot()

    if eid in envs:
        return envs[eid].config


def configure(eid: str, config: Env.Config) -> None:
    """Configures an environment"""

    with State() as envs:
        envs[eid].config = config


def deactivate(*, pid: Optional[int] = None, kernel_id: Optional[str] = None) -> None:
    """Detatches a process or a kernel and deactivates inactive environments"""

    with State() as envs:
        envs.cleanup(
            poll_pid=(lambda pid_: pid_ != pid) if pid else None,
            poll_kernel=(lambda kernel_id_: kernel_id_ != kernel_id)
            if kernel_id
            else None,
        )



from dataclasses import dataclass
import os


class Base:
    def __init__(self, root: str, templates: str) -> None:
        self.root = root
        self.templates = templates

    def _path(self, *paths: str) -> str:
        return os.path.join(self.root, *paths)

    def copy_file(self, *paths: str, **kwargs: object) -> str:
        """
        Copy a file from the inputs directory to the output directory.
        The file contents are formatted with the passed keyword arguments.

        :return: Output file path
        """
        path = self._path(*paths)

        os.makedirs(os.path.dirname(path), exist_ok=True)

        with open(os.path.join(self.templates, *paths), "r") as f:
            contents = f.read()

        if kwargs:
            contents = contents.format(**kwargs)

        with open(path, "w") as f:
            f.write(contents)

        return path

    def publish(self) -> None:
        pass


class Prometheus(Base):
    @dataclass
    class Config:
        exporter_port: int

    def __init__(self, root: str, config: Config) -> None:
        self.config = config

        super().__init__(
            root, os.path.join(os.path.dirname(__file__), "export", "prometheus")
        )

    def publish(self) -> None:
        self.copy_file("prometheus.yml", exporter_port=self.config.exporter_port)


class Grafana(Base):
    @dataclass
    class Config:
        prometheus_url: str = "http://localhost:9090"

    def __init__(self, root: str, config: Config) -> None:
        self.config = config

        super().__init__(
            root, os.path.join(os.path.dirname(__file__), "export", "grafana")
        )

    def publish(self) -> None:
        overview_dashboard = self.copy_file("dashboards", "overview.json")

        self.copy_file(
            "provisioning",
            "datasources",
            "default.yml",
            prometheus_url=self.config.prometheus_url,
        )

        self.copy_file(
            "provisioning",
            "dashboards",
            "default.yml",
            dashboard_dir=self._path("dashboards"),
        )

        self.copy_file(
            "grafana.ini",
            provisioning=self._path("provisioning"),
            default_home_dashboard_path=overview_dashboard,
        )


def publish_config_files(root: str, *, prometheus_exporter_port: int) -> None:
    """
    Publish Prometheus and Grafana configuration files.
    """
    Prometheus(
        os.path.join(root, "prometheus"),
        Prometheus.Config(exporter_port=prometheus_exporter_port),
    ).publish()

    Grafana(
        os.path.join(root, "grafana"),
        Grafana.Config(),
    ).publish()



import argparse

import genv.utils
import genv.core
import genv.sdk


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv deactivate" arguments to a parser.
    """

    pass


def run(shell: int, args: argparse.Namespace) -> None:
    """
    Runs the "genv deactivate" logic.
    """

    if not genv.sdk.active():
        raise RuntimeError("Not running in an active environment")

    with genv.utils.global_lock():
        genv.core.envs.deactivate(pid=shell)

        # clean up devices in case this environment had attached devices and this
        # is the last terminal from the environment
        genv.core.devices.cleanup()

    print(
        f"""
_genv_unset_envs
_genv_restore_envs
"""
    )



import argparse
import asyncio
import dataclasses
import itertools
import os
import shutil
import sys
from typing import Iterable, NoReturn, Optional, Tuple

import genv

QUERIES = {
    "hostname": lambda host, env: host.hostname,
    "eid": lambda host, env: env.eid,
    "creation": lambda host, env: env.creation,
}


async def find_available_host(
    config: genv.remote.Config, gpus: Optional[int]
) -> genv.remote.Host:
    """
    Finds a host with enough available resources.
    Raises 'RuntimeError' if cannot find such host.

    :param gpus: Device count specification

    :return: Host with enough available resources
    """
    # TODO(raz): avoid taking snapshots if not needed when 'gpus' is None
    hosts, snapshots = await genv.remote.core.devices.snapshot(config)

    for host, snapshot in zip(hosts, snapshots):
        if gpus is not None:
            available = len(snapshot.filter(attached=False))

            if available < gpus:
                continue

        return host

    raise RuntimeError("Cannot find a host with enough available resources")


def _exec_ssh(host: genv.remote.Host, command: str) -> NoReturn:
    """Replaces this process with an SSH connection.

    :param host: Remote host to connect to.
    :param command: Command to execute on the remote host.
    """

    path = shutil.which("ssh")

    # NOTE(raz): the '-t' argument is important as it makes the remote process to
    # terminate when the SSH connection gets disconnected (https://superuser.com/a/20708).
    args = [path, "-t", host.hostname, command]

    os.execv(path, args)


async def do_activate(
    config: genv.remote.Config,
    gpus: Optional[int],
    name: Optional[str],
    prompt: bool,
) -> NoReturn:
    """
    Finds a remote host with enough available resources, connects to it using SSH
    and activates an environment inside with the given specification.

    :param gpus: Device count specification
    :param name: Environment name specification
    :param prompt: Change the remote shell prompt

    :return: Does not return
    """
    host = await find_available_host(config, gpus)

    env = {"GENV_PREACTIVATE": 1, "GENV_PREACTIVATE_PROMPT": "1" if prompt else "0"}

    if gpus:
        env.update({"GENV_PREACTIVATE_GPUS": gpus})

    if name:
        env.update({"GENV_PREACTIVATE_ENVIRONMENT_NAME": name})

    path = shutil.which("ssh")
    args = (
        [path]
        + list(itertools.chain(*(["-o", f"SendEnv={variable}"] for variable in env)))
        + [host.hostname]
    )
    env = dict(os.environ, **{k: str(v) for k, v in env.items()})

    os.execve(path, args, env)


async def do_devices(
    config: genv.remote.Config,
    format: str,
    header: bool,
    summary: bool,
) -> None:
    """
    Prints information about devices on multiple hosts.

    :param format: Output format
    :param header: Print a header as part of the output
    :param summary: Print a summary as part of the output

    :return: None
    """
    hosts, snapshots = await genv.remote.core.devices.snapshot(config)

    if header:
        if format == "csv":
            print("HOST,TOTAL,AVAILABLE")
        elif format == "tui":
            print("HOST                     TOTAL    AVAILABLE")

    total = 0
    available = 0

    for host, snapshot in zip(hosts, snapshots):
        host_total = len(snapshot)
        host_available = len(snapshot.filter(attached=False))

        total += host_total
        available += host_available

        if format == "csv":
            print(f"{host.hostname},{host_total},{host_available}")
        elif format == "tui":
            print(f"{host.hostname:<25}{host_total:<9}{host_available}")

    if summary:
        print(
            f"\nTotal {total} devices with {available} available on {len(hosts)} hosts"
        )


async def do_enforce(
    config: genv.remote.Config,
    interval: int,
    non_env_processes: bool,
    env_devices: bool,
    env_memory: bool,
    max_devices_per_user: Optional[int],
    max_devices_for_user: Optional[Iterable[Tuple[str, int]]],
) -> None:
    """
    Enforce GPU usage on multiple hosts.

    :param interval: Interval in seconds; 0 means run once.

    :return: None
    """
    while True:
        hosts, snapshots = await genv.remote.core.snapshot(config)

        surveys = [
            genv.entities.enforce.Survey(snapshot, host.hostname)
            for host, snapshot in zip(hosts, snapshots)
        ]

        if non_env_processes:
            genv.enforce.rules.non_env_processes(*surveys)

        if env_devices:
            genv.enforce.rules.env_devices(*surveys)

        if env_memory:
            genv.enforce.rules.env_memory(*surveys)

        if max_devices_per_user is not None:
            genv.enforce.rules.max_devices_per_user(
                *surveys,
                maximum=max_devices_per_user,
                maximum_for_user=(
                    dict(max_devices_for_user) if max_devices_for_user else {}
                ),
            )

        reports = [survey.report for survey in surveys]

        await genv.remote.enforce.execute(
            config=genv.remote.Config(
                hosts=[host for host, report in zip(hosts, reports) if report],
                throw_on_error=config.throw_on_error,
                quiet=config.quiet,
            ),
            reports=[report for report in reports if report],
        )

        if interval == 0:
            break

        await asyncio.sleep(interval)


async def do_envs(
    config: genv.remote.Config,
    format: str,
    header: bool,
    summary: bool,
    timestamp: bool,
) -> None:
    """
    Prints information about active environments on multiple hosts.

    :param format: Output format
    :param header: Print a header as part of the output
    :param summary: Print a summary as part of the output

    :return: None
    """
    hosts, snapshots = await genv.remote.core.envs.snapshot(config)

    if header:
        if format == "csv":
            print("HOST,ID,USER,NAME,CREATED")
        elif format == "tui":
            print(
                "HOST                     ID      USER            NAME            CREATED"
            )

    for host, snapshot in zip(hosts, snapshots):
        for env in snapshot:
            user = env.username or ""
            name = env.config.name or ""
            created = env.creation if timestamp else env.time_since

            if format == "csv":
                print(f"{host.hostname},{env.eid},{user},{name},{created}")
            elif format == "tui":
                print(f"{host.hostname:<25}{env.eid:<8}{user:<16}{name:<16}{created}")

    if summary:
        print(
            f"\nTotal {sum([len(snapshot) for snapshot in snapshots])} environments on {len(hosts)} hosts"
        )


async def do_llm(config: genv.remote.Config, args: argparse.Namespace):
    """Runs the "genv llm" logic."""

    if args.llm_command == "attach":
        await do_llm_attach(config, args.model)
    elif args.llm_command == "ps":
        await do_llm_ps(config, args.format, args.header, args.summary, args.timestamp)
    elif args.llm_command == "serve":
        await do_llm_serve(config, args.model, args.gpus)
    else:
        await do_llm_ps(config)


async def do_llm_attach(config: genv.remote.Config, model: str) -> NoReturn:
    """
    Attaches to a running LLM on a remote host.
    Raises RuntimeError if could not find the LLM.

    :param model: Model name.
    """

    hosts, snapshots = await genv.remote.core.envs.snapshot(config)

    for host, snapshot in zip(hosts, snapshots):
        replicas = len(snapshot.filter(name=f"llm/{model}"))

        if replicas > 0:
            # TODO(raz): we currently attach to the first node; we should pick a node in a smarter way.
            _exec_ssh(host, f"genv llm attach {model}")

    raise RuntimeError(f"Could not find LLM model '{model}'")


async def do_llm_ps(
    config: genv.remote.Config,
    format: str = "tui",
    header: bool = True,
    summary: bool = True,
    timestamp: bool = False,
) -> None:
    """
    Prints information about active LLMs on remote hosts.

    :param format: Output format
    :param header: Print a header as part of the output
    :param summary: Print a summary as part of the output

    :return: None
    """
    hosts, snapshots = await genv.remote.core.envs.snapshot(config)

    if header:
        if format == "csv":
            print("HOST,MODEL,PORT,CREATED,EID,USER")
        elif format == "tui":
            print(
                "HOST                     MODEL       PORT    CREATED              EID     USER"
            )

    total = 0

    for host, snapshot in zip(hosts, snapshots):
        for env in snapshot:
            if not (env.config.name and env.config.name.startswith("llm/")):
                continue

            total += 1

            model = env.config.name.split("llm/")[1]
            port = "N/A"
            created = env.creation if timestamp else env.time_since
            eid = env.eid
            user = env.username or ""

            if format == "csv":
                print(f"{host.hostname},{model},{port},{created},{eid}{user}")
            elif format == "tui":
                print(
                    f"{host.hostname:<25}{model:<12}{port:<8}{created:<21}{eid:<8}{user}"
                )

    if summary:
        print(f"\nTotal {total} LLMs on {len(hosts)} hosts")


async def do_llm_serve(
    config: genv.remote.Config,
    model: str,
    gpus: Optional[int],
) -> NoReturn:
    """Runs an LLM server in a newly created environment on a remote host."""

    host = await find_available_host(config, gpus)

    command = f"genv llm serve {model}"

    if gpus:
        command = f"{command} --gpus {gpus}"

    _exec_ssh(host, command)


async def do_monitor(
    config: genv.remote.Config, config_dir: str, port: int, interval: int
) -> None:
    """
    Runs a Prometheus exporter in the foreground and publishes Prometheus and Grafana config files.
    """
    try:
        from genv.metrics import publish_config_files, SPECS
        from genv.remote.metrics import Collection
        import prometheus_client
    except ModuleNotFoundError as e:
        if e.name != "prometheus_client":
            raise

        print(f"ERROR: Python package '{e.name}' is required", file=sys.stderr)
        exit(1)

    # https://github.com/prometheus/client_python#disabling-default-collector-metrics
    prometheus_client.REGISTRY.unregister(prometheus_client.GC_COLLECTOR)
    prometheus_client.REGISTRY.unregister(prometheus_client.PLATFORM_COLLECTOR)
    prometheus_client.REGISTRY.unregister(prometheus_client.PROCESS_COLLECTOR)

    prometheus_client.start_http_server(port)

    publish_config_files(config_dir, prometheus_exporter_port=port)

    collection = Collection(SPECS)

    while True:
        hosts, systems = await genv.remote.core.system(config)

        hosts_with_genv = [
            host for host, system in zip(hosts, systems) if system.genv.installed
        ]

        hosts_with_snapshots, snapshots = await genv.remote.core.snapshot(
            dataclasses.replace(config, hosts=hosts_with_genv)
        )

        hostname_to_snapshot = {
            host.hostname: snapshot
            for host, snapshot in zip(hosts_with_snapshots, snapshots)
        }

        # inflate snapshots to match hosts
        snapshots = [hostname_to_snapshot.get(host.hostname, None) for host in hosts]

        collection.cleanup(hosts, systems, snapshots)
        collection.update(hosts, systems, snapshots)

        await asyncio.sleep(interval)


async def do_query(
    config: genv.remote.Config, name: str, queries: Iterable[str]
) -> None:
    """
    Query environment data using a query string and print the results to the string.

    :param name: Environment name
    :param queries: List of query names

    :return: None
    """
    hosts, snapshots = await genv.remote.core.envs.snapshot(config)

    for host, snapshot in zip(hosts, snapshots):
        for env in snapshot:
            if env.config.name != name:
                continue

            def query(name: str) -> str:
                result = QUERIES[name](host, env)

                return "" if result is None else str(result)

            print(",".join(query(name) for name in queries))


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv remote" arguments to a parser.
    """

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-H",
        "-host",
        "--host",
        dest="hostnames",
        help="Comma-separated hostnames or IP addresses",
    )
    group.add_argument(
        "-hostfile",
        "--hostfile",
        help="A file containing one hostname or IP address per line",
    )

    parser.add_argument(
        "-l",
        "--username",
        help="SSH login name",
    )

    parser.add_argument(
        "-t",
        "--timeout",
        type=int,
        help="SSH connection timeout",
    )

    parser.add_argument(
        "-e",
        "--exit-on-error",
        dest="throw_on_error",
        action="store_true",
        help="Exit on SSH error to one or more hosts (default: %(default)s)",
    )

    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Ignore SSH errors (default: %(default)s)",
    )

    subparsers = parser.add_subparsers(dest="command")

    def activate(parser):
        parser.add_argument("--name", help="Environment name")
        parser.add_argument("--gpus", type=int, help="Environment device count")
        parser.add_argument(
            "--no-prompt",
            dest="prompt",
            action="store_false",
            help="Don't change the shell prompt",
        )

    def devices(parser):
        parser.add_argument(
            "--no-header",
            dest="header",
            action="store_false",
            help="Do not print column headers",
        )
        parser.add_argument(
            "--no-summary",
            dest="summary",
            action="store_false",
            help="Do not print summary",
        )
        parser.add_argument(
            "--format",
            choices=["csv", "tui"],
            help="Output format; CSV or TUI (Text-based user interface)",
            default="tui",
        )

    def enforce(parser):
        parser.add_argument(
            "--interval",
            type=int,
            default=10,
            help="interval in seconds between enforcement cycles; 0 means run once (default: %(default)s)",
        )

        enforcements = parser.add_argument_group("enforcements")

        def add_enforcement(
            name: str, *, dest: str, help: str, default: bool = True
        ) -> None:
            """
            Adds a mutual exclusive group for an enforcement flag with true and false actions.
            """
            group = enforcements.add_mutually_exclusive_group()

            group.add_argument(
                f"--{name}",
                action="store_true",
                default=default,
                help=help + " (default: %(default)s)",
                dest=dest,
            )

            group.add_argument(
                f"--no-{name}",
                action="store_false",
                dest=dest,
            )

        add_enforcement(
            "non-env-processes",
            dest="non_env_processes",
            help="terminate processes that are not running in a GPU environment",
            default=False,
        )

        add_enforcement(
            "env-devices",
            dest="env_devices",
            help="enforce environment attached devices",
            default=True,
        )

        add_enforcement(
            "env-memory",
            dest="env_memory",
            help="enforce environment memory capacity if set",
            default=True,
        )

        enforcements.add_argument(
            "--max-devices-per-user",
            type=int,
            help="maximum allowed attached devices for each user",
        )

        def max_devices_for_user(value: str):
            try:
                username, maximum = value.split("=")

                return username, int(maximum)
            except (ValueError, SyntaxError):
                raise argparse.ArgumentTypeError(f"not a valid spec: {value}")

        enforcements.add_argument(
            "--max-devices-for-user",
            nargs="+",
            help="per-user specification of maximum allowed attached devices",
            metavar="username=maximum",
            type=max_devices_for_user,
        )

    def envs(parser):
        parser.add_argument(
            "--no-header",
            dest="header",
            action="store_false",
            help="Do not print column headers",
        )
        parser.add_argument(
            "--no-summary",
            dest="summary",
            action="store_false",
            help="Do not print summary",
        )
        parser.add_argument(
            "--timestamp", action="store_true", help="Print a non-prettified timestamp"
        )
        parser.add_argument(
            "--format",
            choices=["csv", "tui"],
            help="Output format; CSV or TUI (Text-based user interface)",
            default="tui",
        )

    def llm(parser):
        subparsers = parser.add_subparsers(dest="llm_command")

        def attach(parser):
            parser.add_argument("model")

        def ps(parser):
            parser.add_argument(
                "--no-header",
                dest="header",
                action="store_false",
                help="Do not print column headers",
            )
            parser.add_argument(
                "--no-summary",
                dest="summary",
                action="store_false",
                help="Do not print summary",
            )
            parser.add_argument(
                "--timestamp",
                action="store_true",
                help="Print a non-prettified timestamp",
            )
            parser.add_argument(
                "--format",
                choices=["csv", "tui"],
                help="Output format; CSV or TUI (Text-based user interface)",
                default="tui",
            )

        def serve(parser):
            parser.add_argument("model")
            parser.add_argument("--gpus", type=int, help="Device count")

        for command, help in [
            (attach, "Attach to a running LLM model"),
            (ps, "Print information about active LLMs"),
            (serve, "Start an LLM server"),
        ]:
            command(subparsers.add_parser(command.__name__, help=help))

    def monitor(parser):
        parser.add_argument(
            "--config-dir",
            default=genv.utils.get_temp_file_path("metrics"),
            help="Directory to create Prometheus and Grafana config files at (default: %(default)s)",
        )

        parser.add_argument(
            "-p",
            "--port",
            type=int,
            default=8000,
            help="Port for Prometheus exporter to listen on (default: %(default)s)",
        )

        parser.add_argument(
            "-i",
            "--interval",
            type=int,
            default=10,
            help="Interval in seconds between collections (default: %(default)s)",
        )

    def query(parser):
        parser.add_argument("--name", required=True, help="Environment name")
        parser.add_argument(
            "--query",
            "--queries",
            dest="queries",
            nargs="+",
            choices=QUERIES.keys(),
            required=True,
        )

    for command, help in [
        (activate, "Activate environment in a remote host"),
        (devices, "Show device information from all hosts"),
        (enforce, "Enforce GPU usage on all hosts"),
        (envs, "List all active environments"),
        (llm, "Run and attach to LLMs on remote hosts"),
        (monitor, "Monitor remote hosts using Prometheus and Grafana"),
        (query, "Query environments or a specific one"),
    ]:
        submodule = command.__name__
        aliases = [f"{submodule}s"] if submodule in ["llm"] else []

        command(subparsers.add_parser(submodule, aliases=aliases, help=help))


def parse_hosts(args: argparse.Namespace) -> Iterable[genv.remote.Host]:
    """Parses arguments and returns a list of hosts."""

    if args.hostfile:
        with open(args.hostfile, "r") as f:
            hostnames = [
                line
                for line in [line.strip() for line in f.readlines()]
                if line and not line.startswith("#")
            ]
    else:
        hostnames = args.hostnames.split(",")

    def parse_host(hostname: str) -> genv.remote.Host:
        username, hostname = (
            hostname.split("@") if "@" in hostname else (args.username, hostname)
        )

        return genv.remote.Host(hostname, username, args.timeout)

    return [parse_host(hostname) for hostname in hostnames]


async def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv remote" logic.
    """

    hosts = parse_hosts(args)
    config = genv.remote.Config(hosts, args.throw_on_error, args.quiet)

    if args.command == "activate":
        await do_activate(config, args.gpus, args.name, args.prompt)
    elif args.command == "devices":
        await do_devices(config, args.format, args.header, args.summary)
    elif args.command == "enforce":
        await do_enforce(
            config,
            args.interval,
            args.non_env_processes,
            args.env_devices,
            args.env_memory,
            args.max_devices_per_user,
            args.max_devices_for_user,
        )
    elif args.command == "envs":
        await do_envs(
            config,
            args.format,
            args.header,
            args.summary,
            args.timestamp,
        )
    elif args.command in ["llm", "llms"]:
        await do_llm(config, args)
    elif args.command == "monitor":
        await do_monitor(config, args.config_dir, args.port, args.interval)
    elif args.command == "query":
        await do_query(config, args.name, args.queries)



import asyncio

from genv.entities import Process, Processes

import genv.utils.nvidia_smi


async def snapshot() -> Processes:
    """
    Returns a snapshot of all running compute processes.
    """
    uuids, apps = await asyncio.gather(
        genv.utils.nvidia_smi.device_uuids(), genv.utils.nvidia_smi.compute_apps()
    )

    pid_to_apps = {
        pid: [app for app in apps if app["pid"] == pid]
        for pid in set(app["pid"] for app in apps)
    }

    return Processes(
        [
            Process(
                pid=pid,
                used_gpu_memory=[
                    Process.Usage(
                        index=uuids[app["gpu_uuid"]], gpu_memory=app["used_gpu_memory"]
                    )
                    for app in apps
                ],
                eid=Process.eid(pid),
            )
            for pid, apps in pid_to_apps.items()
        ]
    )



import json
from typing import Any, Dict

from genv.entities import Device, Devices, Env, Envs, Process, Processes, Snapshot, Report

# TODO(raz): test here that all types have a different set of keys for creation
Types = [
    Device,
    Device.Attachement,
    Devices,
    Env,
    Env.Config,
    Envs,
    Process,
    Process.Usage,
    Processes,
    Report,
    Snapshot,
]


class JSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Dict:
        if o.__class__ in Types:
            return o.__dict__

        return super().default(o)


class JSONDecoder(json.JSONDecoder):
    def __init__(self) -> None:
        super().__init__(object_hook=self.object_hook)

    def object_hook(self, d: Dict) -> Any:
        # all classes are Python dataclasses and therefore have a constructor that receive all members as arguments.
        # we rely on `co_varnames` to find them dynamically.
        # here's the documentation: https://docs.python.org/3.7/reference/datamodel.html#index-55.
        for cls in Types:
            varnames = cls.__init__.__code__.co_varnames[1:]  # ignore 'self'

            if set(varnames) == set(d.keys()):
                o = cls(*(d[varname] for varname in varnames))

                # NOTE(raz): serializing the class 'Report' is not properly supported because it has
                # a dictionary field 'detach' which its keys are integers.
                # these integers are converted to strings when serializing to JSON.
                # therefore, we have to revert this conversion manuallu here.
                if cls == Report:
                    # TODO(raz): support this properly by refactoring 'Report'
                    o.detach = {int(index): envs for index, envs in o.detach.items()}

                return o

        return d



import argparse
from typing import Iterable, Optional

import genv

QUERIES = {
    "eid": lambda env: env.eid,
    "creation": lambda env: env.creation,
    "uid": lambda env: env.uid,
    "username": lambda env: env.username,
    "config.name": lambda env: env.config.name,
    "config.gpus": lambda env: env.config.gpus,
    "config.gpu_memory": lambda env: env.config.gpu_memory,
    "pids": lambda env: " ".join(str(pid) for pid in env.pids),
    "kernel_ids": lambda env: " ".join(kernel_id for kernel_id in env.kernel_ids),
}


def do_activate(
    snapshot: genv.Envs,
    eid: str,
    uid: int,
    username: Optional[str],
    pid: Optional[int],
    kernel_id: Optional[str],
) -> None:
    """
    Activates an environment.
    Creates the environment if does not exist already.
    Attaches the given process id or and jupyter kernel ids if exist.
    """
    if eid not in snapshot:
        snapshot.activate(
            eid=eid,
            uid=uid,
            username=username,
        )

    snapshot[eid].attach(pid=pid, kernel_id=kernel_id)


def do_config(
    snapshot: genv.Envs, eid: str, command: str, args: argparse.Namespace
) -> None:
    """
    Update environment config variables - either set a specific variable or clear it.

    :param command: a string describing which variable in the environment config to change
    :param args: parameters for the config change.
                For example - gpus count variable, clear/not clear config variable, etc.
    :return: None
    """
    if eid in snapshot:
        env = snapshot[eid]

        if command == "gpus":
            env.config.gpus = args.count
        elif command == "name":
            env.config.name = args.name
        elif command == "gpu-memory":
            env.config.gpu_memory = args.gpu_memory


def do_deactivate(snapshot: genv.Envs, pid: int) -> None:
    """
    Detaches the given process identifier.
    """
    snapshot.cleanup(poll_pid=lambda pid_: pid_ != pid)


def do_find(snapshot: genv.Envs, pid: Optional[int], kernel_id: Optional[str]) -> None:
    """
    Prints the environment identifiers of the given process or kernel.
    """
    for env in snapshot.find(pid=pid, kernel_id=kernel_id):
        print(env.eid)


def do_ps(snapshot: genv.Envs, format: str, header: bool, timestamp: bool) -> None:
    """
    Print environments data in a human-readable fashion to the stdio.

    :param format: Which format should be used to print the environment data - csv or tui (default: tui).
    :param header: a flag instructing if we should print a header as part of the output (default: print header).
    :param timestamp: a flag instructing if we should print creation time as is or relative duration (now - creation).
                    (default: relative duration)
    :return: None
    """
    if header:
        if format == "csv":
            print("ID,USER,NAME,CREATED,PID(S)")
        elif format == "tui":
            print("ID      USER            NAME            CREATED              PID(S)")

    for env in snapshot:
        eid = env.eid
        user = f"{env.username}({env.uid})" if env.username else env.uid
        name = env.config.name or ""
        created = env.creation if timestamp else genv.utils.time_since(env.creation)
        pids = " ".join(str(pid) for pid in env.pids)
        # TODO(raz): print kernel ids

        if format == "csv":
            print(f"{eid},{user},{name},{created},{pids}")
        elif format == "tui":
            print(f"{eid:<8}{user:<16}{name:<16}{created:<21}{pids}")


def do_query(snapshot: genv.Envs, eid: Optional[str], queries: Iterable[str]) -> None:
    """
    Queries environment data using a query string. The queries results will be printed into the stdio.
    """
    eids = [eid] if eid is not None else snapshot.eids

    for eid in eids:
        if eid not in snapshot:
            continue

        env = snapshot[eid]

        def query(name: str) -> str:
            result = QUERIES[name](env)

            return "" if result is None else str(result)

        print(",".join(query(name) for name in queries))


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv envs" arguments to a parser.
    """

    parser.add_argument(
        "--no-cleanup",
        dest="cleanup",
        action="store_false",
        help="Do not perform clean up",
    )

    parser.add_argument("--reset", action="store_true", help="Reset previous state")

    subparsers = parser.add_subparsers(dest="command")

    def activate(parser):
        parser.add_argument("--eid", required=True, help="Environment identifier")
        parser.add_argument(
            "--uid", type=int, required=True, help="User identifier"
        )  # TODO(raz): should we make this optional?
        parser.add_argument("--username", help="User name")

        group = parser.add_mutually_exclusive_group(required=True)

        group.add_argument("--pid", type=int, help="Process identifier")
        group.add_argument("--kernel-id", help="Jupyter kernel identifier")

    def config(parser):
        def gpus(parser):
            group = parser.add_mutually_exclusive_group(required=True)
            group.add_argument("--count", type=int, help="Device count")
            group.add_argument(
                "--clear", action="store_true", help="Clear device count"
            )

        def name(parser):
            group = parser.add_mutually_exclusive_group(required=True)
            group.add_argument("--name", help="Environment name")
            group.add_argument(
                "--clear", action="store_true", help="Clear environment name"
            )

        def gpu_memory(parser):
            group = parser.add_mutually_exclusive_group(required=True)
            group.add_argument("--gpu-memory", help="Environment GPU memory capacity")
            group.add_argument(
                "--clear",
                action="store_true",
                help="Clear environment GPU memory capacity",
            )

        parser.add_argument("--eid", required=True, help="Environment identifier")

        subparsers = parser.add_subparsers(dest="config", required=True)

        for option in [
            (gpus, "Configure device count for environment"),
            (name, "Configure environment name"),
            (gpu_memory, "gpu-memory", "Configure environment GPU memory capacity"),
        ]:
            if len(option) == 2:
                command, help = option
                name_ = command.__name__
            else:
                command, name_, help = option

            command(subparsers.add_parser(name_, help=help))

    def deactivate(parser):
        parser.add_argument("--pid", type=int, required=True, help="Process identifier")

    def find(parser):
        group = parser.add_mutually_exclusive_group(required=True)

        group.add_argument("--pid", type=int, help="Process identifier")
        group.add_argument("--kernel-id", help="Jupyter kernel identifier")

    def ps(parser):
        parser.add_argument(
            "--no-header",
            dest="header",
            action="store_false",
            help="Do not print column headers",
        )
        parser.add_argument(
            "--timestamp", action="store_true", help="Print a non-prettified timestamp"
        )
        parser.add_argument(
            "--format",
            choices=["csv", "tui"],
            help="Output format; CSV or TUI (Text-based user interface)",
            default="tui",
        )

    def query(parser):
        parser.add_argument("--eid", help="Environment identifier")
        parser.add_argument(
            "--query",
            "--queries",
            dest="queries",
            nargs="+",
            choices=QUERIES.keys(),
            required=True,
        )

    for command, help in [
        (activate, "Activate a process or a Jupyter kernel"),
        (config, "Configure an environment"),
        (deactivate, "Deactivate a process or a Jupyter kernel"),
        (find, "Find the environment of a process or a Jupyter kernel"),
        (ps, "Print information about active environments"),
        (query, "Query environments or a specific one"),
    ]:
        command(subparsers.add_parser(command.__name__, help=help))


def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv envs" logic.
    """

    with genv.utils.global_lock():
        with genv.core.envs.State(args.cleanup, args.reset) as envs:
            if args.command == "activate":
                do_activate(
                    envs,
                    args.eid,
                    args.uid,
                    args.username,
                    args.pid,
                    args.kernel_id,
                )
            elif args.command == "config":
                do_config(envs, args.eid, command=args.config, args=args)
            elif args.command == "deactivate":
                do_deactivate(envs, args.pid)
            elif args.command == "find":
                do_find(envs, args.pid, args.kernel_id)
            elif args.command == "ps":
                do_ps(envs, args.format, args.header, args.timestamp)
            elif args.command == "query":
                do_query(envs, args.eid, args.queries)
            else:
                do_ps(envs, format="tui", header=True, timestamp=False)



import dataclasses
from typing import Iterable, Optional

from genv.entities import Snapshot, System
from genv.metrics import Collection as Base, Spec

from ..utils import Host


class Collection(Base):
    """
    A hostname aware metric collection.
    """

    def __init__(self, specs: Iterable[Spec]) -> None:
        return super().__init__(
            [
                dataclasses.replace(spec, labelnames=("hostname",) + spec.labelnames)
                for spec in specs
            ]
        )

    def cleanup(
        self,
        hosts: Iterable[Host],
        systems: Iterable[Optional[System]],
        snapshots: Iterable[Optional[Snapshot]],
    ) -> None:
        """
        Cleans up metric label values.
        """
        hostnames = [host.hostname for host in hosts]

        for metric in self:
            for labelvalues in metric.label_sets():
                if labelvalues[0] not in hostnames:
                    metric.remove(*labelvalues)

        for host, system, snapshot in zip(hosts, systems, snapshots):
            super().cleanup(system, snapshot, header=host.hostname)

    def update(
        self,
        hosts: Iterable[Host],
        systems: Iterable[Optional[System]],
        snapshots: Iterable[Optional[Snapshot]],
    ) -> None:
        """
        Updates metrics of hosts according to the given data.
        """
        for host, system, snapshot in zip(hosts, systems, snapshots):
            super().update(system, snapshot, labels={"hostname": host.hostname})



from contextlib import ExitStack, contextmanager
import subprocess
from typing import Any, Iterable, Optional, Union

import genv.utils
from genv.entities import Device, Devices, Env
import genv.serialization

import genv.core.envs

from .utils import File


class State(File[Devices]):
    def __init__(self, cleanup: bool = True, reset: bool = False) -> None:
        super().__init__(
            genv.utils.get_temp_file_path("devices.json"),
            cleanup,
            reset,
        )

    def _create(self):
        # move to genv.utils.nvidia_smi once async is supported here
        devices_total_memory = [
            f"{int(line)}mi"
            for line in subprocess.check_output(
                "GENV_BYPASS=1 nvidia-smi --query-gpu=memory.total --format=csv,noheader,nounits",
                shell=True,
            ).splitlines()
        ]

        return Devices(
            [
                Device(index, total_memory, [])
                for index, total_memory in enumerate(devices_total_memory)
            ]
        )

    def _convert(self, o: Union[Any, Devices]) -> Devices:
        # the following logic converts state files from versions <= 0.8.0.
        # note that the indicator is 'o.devices' and not 'o' itself because the structure of
        # the state file from these versions was similar to the snapshot structure (a single
        # key named "devices") so the JSON decoder parsed the object as a snapshot object.
        if isinstance(o.devices, dict):
            o = Devices(
                [
                    Device(
                        int(index),
                        device["total_memory"],
                        [
                            Device.Attachement(
                                env["eid"],
                                env.get("gpu_memory", None),
                                env["attached"],
                            )
                            for env in device["eids"].values()
                        ],
                    )
                    for index, device in o.devices.items()
                ]
            )

        return o

    def _clean(self, devices: Devices) -> None:
        envs = genv.core.envs.snapshot()

        devices.cleanup(poll_eid=lambda eid: eid in envs.eids)


def cleanup() -> None:
    """Cleans up the state"""

    with State(cleanup=True):
        pass


def snapshot() -> Devices:
    """
    Returns an environments snapshot.
    """
    return State().load()


def attach(
    eid: str,
    *,
    index: Optional[int] = None,
    gpus: Optional[int] = None,
    gpu_memory: Optional[str] = None,
    allow_over_subscription: bool = False,
) -> Iterable[int]:
    """Attaches an environment to devices.

    Does not detach devices if already attached to more devices.

    :return: Attached device indices
    """
    if gpus is not None and index is not None:
        raise ValueError(
            'Can\'t use both "gpus" and "index" in genv.core.devices.attach()'
        )

    with State() as devices:
        env_devices = devices.filter(eid=eid)

        if index is not None:
            if index not in env_devices.indices:
                if not allow_over_subscription and not devices[index].available(
                    gpu_memory
                ):
                    raise RuntimeError(f"Device {index} is not available")

                devices.attach(eid, index, gpu_memory)
        elif gpus is not None:
            diff = gpus - len(env_devices)

            if diff > 0:
                not_env_devices = devices.filter(not_indices=env_devices.indices)

                indices = not_env_devices.find_available_devices(
                    diff, gpu_memory, allow_over_subscription
                )

                devices.attach(eid, indices, gpu_memory)
            elif diff < 0:
                pass  # TODO(raz): support detaching devices if already attached to more

        return devices.filter(eid=eid).indices


def detach(eid: str, index: Optional[int] = None) -> Iterable[int]:
    """
    Detaches an environment from a device.
    """
    with State() as devices:
        devices.detach(eid, index)

        return devices.filter(eid=eid).indices


def attached(eid: str) -> Iterable[int]:
    """Returns the indices of devices that are attached to an environment"""

    devices = snapshot()

    return devices.filter(eid=eid).indices


def get_lock_path(index: int, create: bool = False) -> str:
    """
    Returns the path of a device lock file.
    Creates the file if requested and it does not exist.
    """

    path = genv.utils.get_temp_file_path(f"devices/{index}.lock")

    if create:
        genv.utils.create_lock(path)

    return path


@contextmanager
def lock(indices: Union[Iterable[int], int]) -> None:
    """
    Obtain exclusive access to a device or a few devices.
    """
    if isinstance(indices, int):
        indices = [indices]

    # we sort here to decrease the chance of deadlock when different processes
    # try to lock the same devices in a different order.
    indices = sorted(indices)

    # NOTE(raz): currently, we wait on a lock even if it is already taken by our environment.
    # we should think if this is the desired behavior and if it's possible to lock once per environment.

    with ExitStack() as es:
        for index in indices:
            es.enter_context(genv.utils.access_lock(get_lock_path(index)))

        yield

    # TODO(raz): currently we wait for the entire device to become available.
    # we should support fractional usage and allow multiple environments to
    # access the device if the sum of their memory capacity fits.



import math
import os
import pynvml
import ray

import genv


def remote(**options):
    """Wraps a Ray remote function with a Genv environment.

    Inspired by https://github.com/ray-project/ray/blob/ray-2.5.1/python/ray/_private/worker.py#L3059
    """

    num_gpus = options.get("num_gpus", None)

    if num_gpus is None:
        raise ValueError(
            "The @genv.ray.remote decorator must be applied only when the argument 'num_gpus' is being used."
        )

    def _decorator(function):
        @ray.remote(**options)
        def _wrapper(*args, **kwargs):
            indices = ray.get_gpu_ids()

            config = genv.sdk.Env.Config(
                name=f"ray/{os.getpid()}", gpus=math.ceil(num_gpus)
            )

            if 0 < num_gpus < 1:
                if "GENV_MOCK_DEVICE_TOTAL_MEMORY" in os.environ:
                    total_bytes = genv.utils.memory_to_bytes(
                        os.environ["GENV_MOCK_DEVICE_TOTAL_MEMORY"]
                    )
                else:
                    index = indices[0]

                    pynvml.nvmlInit()
                    handle = pynvml.nvmlDeviceGetHandleByIndex(index)
                    memory_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
                    total_bytes = memory_info.total

                config.gpu_memory = str(math.floor(total_bytes * num_gpus))

            with genv.sdk.activate():
                genv.sdk.configure(config)

                # TODO(raz): support attaching to multiple indices at once
                for index in indices:
                    genv.sdk.attach(index=index, allow_over_subscription=True)

                return function(*args, **kwargs)

        return _wrapper

    return _decorator



from contextlib import contextmanager
import getpass
import os
from pathlib import Path
from typing import Optional

import genv.utils
from genv.entities import Env
import genv.core

from . import devices
from .utils import set_temp_env_var, temp_env_vars, unset_temp_env_var


def eid() -> Optional[str]:
    """Returns the current environment identifier or None if not running in one"""

    return os.environ.get("GENV_ENVIRONMENT_ID")


def active() -> bool:
    """Returns whether running in an active environment"""

    return eid() is not None


def home() -> Optional[Path]:
    """Returns the configuration directory for this environment if exists"""

    for alt in [Path.cwd(), Path.home()]:
        path = alt.joinpath(".genv")

        if path.is_dir():
            return path


def _update_env(config: Env.Config) -> None:
    """Updates the configuration environment variables"""

    for name, value in [
        ("GENV_ENVIRONMENT_NAME", config.name),
        ("GENV_GPU_MEMORY", config.gpu_memory),
        ("GENV_GPUS", config.gpus),
    ]:
        if value is not None:
            set_temp_env_var(name, value)
        else:
            unset_temp_env_var(name)


def configure(config: Env.Config) -> None:
    """Configures the current environment"""

    if not active():
        raise RuntimeError("Not running in an active environment")

    with genv.utils.global_lock():
        genv.core.envs.configure(eid(), config)

    _update_env(config)


def configuration() -> Env.Config:
    """Returns the current environment configuration.

    Raises RuntimeError if not running in an active environment.
    """

    if not active():
        raise RuntimeError("Not running in an active environment")

    name = os.environ.get("GENV_ENVIRONMENT_NAME")
    gpu_memory = os.environ.get("GENV_GPU_MEMORY")
    gpus = os.environ.get("GENV_GPUS")

    return Env.Config(
        name=name, gpu_memory=gpu_memory, gpus=int(gpus) if gpus else None
    )


def refresh_configuration() -> Env.Config:
    """Refreshes the environment configuration.

    Raises RuntimeError if not running in an active environment.
    """

    if not active():
        raise RuntimeError("Not running in an active environment")

    with genv.utils.global_lock():
        config = genv.core.envs.configuration(eid())

    _update_env(config)

    return config


def load_configuration() -> Env.Config:
    """Loads configuration from disk.

    Raises RuntimeError if not running in an active environment.
    """

    if not active():
        raise RuntimeError("Not running in an active environment")

    config = Env.Config()

    home_dir = home()

    if home_dir:
        config.load(home_dir)

    _update_env(config)

    return config


def save_configuration() -> None:
    """Saves the current configuration to disk.

    Raises RuntimeError if not running in an active environment.
    """

    home_dir = home()

    if home_dir is not None:
        config = configuration()

        config.save(home_dir)


@contextmanager
def activate(*, eid: Optional[str] = None, config: Optional[Env.Config] = None) -> None:
    """A context manager for activating an environment for the current process.

    Configures the environment if configuration is specified.
    Attaches devices if device count is configured.

    Raises RuntimeError if already running in an active environment.

    :param eid: Environment identifier
    :param config: Environment configuration
    """

    if active():
        raise RuntimeError("Already running in an active environment")

    pid = os.getpid()
    eid = eid or str(pid)

    with temp_env_vars():
        set_temp_env_var("GENV_PYTHON", "1")
        set_temp_env_var("GENV_ENVIRONMENT_ID", eid)

        with genv.utils.global_lock():
            genv.core.envs.activate(
                eid, uid=os.getuid(), username=getpass.getuser(), pid=pid
            )

        if config is not None:
            configure(config)
        else:
            config = refresh_configuration()

        indices = devices.refresh_attached()

        if not indices:
            devices.attach()

        try:
            yield
        finally:
            with genv.utils.global_lock():
                genv.core.envs.deactivate(pid=pid)



import errno
import os
import subprocess


def poll_pid(pid: int) -> bool:
    """
    Kill the process of the given pid.
    :param pid: pid of the process to kill
    :return: True iff the process has been killed successfully.
    """
    try:
        os.kill(pid, 0)
    except OSError as e:
        if e.errno == errno.ESRCH:
            return False
        elif e.errno == errno.EPERM:
            return True
        else:
            raise
    else:
        return True


def poll_jupyter_kernel(kernel_id: str) -> bool:
    # TODO(raz): what about the case when 'jupyter' is not available in the
    #            environment that we are currently running in?
    #
    # should we ignore such cases and _not_ cleanup kernels if we don't have the 'jupyter' command?
    # should we look for kernel processes similarly to 'ps -ef | grep kernel-'?
    # should we document the kernel json path when activating a kernel, so that the path will
    # be known in other environments as well? what if we don't have read permissions?

    result = subprocess.run(
        ["sh", "-c", f"ls $(jupyter --runtime-dir)/kernel-{kernel_id}.json"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    return result.returncode == 0



import os
from genv.entities import System

from genv.utils.runners import Local


# TODO(raz): this should be combined with genv.remote.core.system()
async def system() -> System:
    """Returns information about the system"""

    GENV_COMMAND = "command -v genv 2> /dev/null || echo none"
    DEVICES_COMMAND = "nvidia-smi --query-gpu=index,utilization.gpu,temperature.gpu,memory.used,memory.total --format=csv,noheader,nounits"

    command = [
        "bash",
        "-c",
        f"({GENV_COMMAND}) && ({DEVICES_COMMAND})",  # NOTE(raz): when using a local runner we don't need to escape this argument
    ]

    runner = Local({"GENV_BYPASS": "1", **os.environ})
    stdout = (await runner.run(*command, check=True)).stdout

    genv_info, device_infos = stdout.split("\n", 1)

    def parse_genv_info(info: str) -> System.Genv:
        return System.Genv(installed=info != "none")

    def parse_device_info(line: str) -> System.Device:
        index, utilization, temperature, used_memory, total_memory = line.split(", ")

        return System.Device(
            index=int(index),
            utilization=int(utilization),
            temperature=int(temperature),
            used_memory=f"{used_memory}mi",
            total_memory=f"{total_memory}mi",
        )

    return System(
        genv=parse_genv_info(genv_info),
        devices=[parse_device_info(line) for line in device_infos.splitlines()],
    )



from genv.entities.enforce import Survey


def non_env_processes(*surveys: Survey) -> None:
    """
    Terminates processes that are not running in environments.
    """
    for survey in surveys:
        for process in survey.snapshot.processes:
            if process.eid is not None:
                # TODO(raz): should we make sure that the environment actually exists in the snapshot?
                continue

            print(
                f"{f'[{survey.hostname}] ' if survey.hostname else ''}"
                f"Process {process.pid} is not running in a GPU environment"
            )

            survey.terminate(process.pid)



from genv.entities import Snapshot

import genv.core.envs
import genv.core.devices
import genv.core.processes


async def snapshot() -> Snapshot:
    """
    Returns a full system snapshot.
    """

    processes = await genv.core.processes.snapshot()
    envs = genv.core.envs.snapshot()
    devices = genv.core.devices.snapshot()

    return Snapshot(processes, envs, devices)



from dataclasses import dataclass
from typing import Any, Callable, Iterable, Optional, Tuple

from genv.entities import Snapshot

from .type import Type


@dataclass
class Spec:
    """
    A metric specification.

    :param type: Metric group type.
    :param name: Metric name.
    :param documentation: A short description of the metric.
    :param labelnames: Metric label names.
    :param convert: A lambda converting the input data to a metric value; input data type varies per metric.
    :param filter: A lambda for cleaning up expired label names according to the given input data; input data type varies per metric.
    """

    type: Type
    name: str
    documentation: str
    labelnames: Tuple[str] = ()
    convert: Optional[Callable[[Any], float]] = None
    filter: Optional[Callable[[Iterable[str], Snapshot], bool]] = None



import json
from typing import Iterable, Optional, Tuple

from genv.entities import Snapshot
from genv.serialization import JSONDecoder

from ..utils import run, Host, Config, Command

# TODO(raz): should we support cases where 'sudo' is not an option?


async def exec(config: Config, *, type: Optional[str], sudo: bool):
    args = ["usage", "snapshot"]

    if type:
        args.append(f"--type {type}")

    command = Command(args, sudo)

    hosts, stdouts = await run(config, command)

    return hosts, [json.loads(stdout, cls=JSONDecoder) for stdout in stdouts]


async def snapshot(config: Config) -> Tuple[Iterable[Host], Iterable[Snapshot]]:
    """
    Takes usage snapshots on multiple hosts.

    :return: Returns the hosts that succeeded and their snapshots
    """
    return await exec(config, type=None, sudo=True)



from contextlib import contextmanager
import fcntl
import os
from pathlib import Path
import platform
import signal
import subprocess
from typing import Dict, Iterable

import psutil


@contextmanager
def Umask(value: int = 0):
    prev = os.umask(value)
    try:
        yield
    finally:
        os.umask(prev)


class Flock:
    def __init__(self, path: str, mode: int):
        self._path = path
        self._mode = mode

    def __enter__(self):
        self._fd = os.open(self._path, os.O_RDWR | os.O_CREAT | os.O_TRUNC, self._mode)

        try:
            fcntl.flock(self._fd, fcntl.LOCK_EX)
        except (IOError, OSError):
            os.close(self._fd)
            raise

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            fcntl.flock(self._fd, fcntl.LOCK_UN)
        finally:
            os.close(self._fd)


@contextmanager
def access_lock(path: str) -> None:
    """
    Locks an exclusive lock.
    Creates the lock file and its parent directories if not exists.
    """
    with Umask(0):
        Path(path).parent.mkdir(parents=True, exist_ok=True, mode=0o777)

        with Flock(path, mode=0o666):
            yield


def create_lock(path: str) -> None:
    """
    Creates a lock file and its parent directories if not exists.
    """
    with access_lock(path):
        pass  # TODO(raz): create the lock without waiting on it


def get_process_environ(pid: int) -> Dict[str, str]:
    """
    Returns the environment variables of the process with the given identifier.

    Raises 'NotImplementedError' if running in a non-Linux platform as it relies on the Linux proc filesystem.
    Raises 'FileNotFoundError' if no such process.
    Raises 'PermissionError' if there is no sufficient permissions.

    """
    # TODO(raz): support this with psutil.Process().eniron() probably
    if platform.system() != "Linux":
        raise NotImplementedError(
            "genv.utils.os_.get_process_environ is not supported in platforms other than Linux"
        )

    with open(f"/proc/{pid}/environ", "r") as f:
        return {
            variable: value
            for variable, value in (
                line.split("=", 1) for line in f.read().split("\x00") if line
            )
        }


def get_process_listen_ports(pid: int) -> Iterable[int]:
    """Returns the port number on which a process listens."""

    process = psutil.Process(pid)

    connections = [
        connection
        for connection in process.connections()
        if connection.status == psutil.CONN_LISTEN
    ]

    return [connection.laddr.port for connection in connections]


def pgrep(name: str) -> Iterable[int]:
    """
    Returns the identifiers of processes with the given name.
    """
    return [
        int(pid)
        for pid in subprocess.check_output(
            f"pgrep {name} || true", shell=True
        ).splitlines()
    ]


def cmdline(pid: int) -> Iterable[str]:
    """
    Returns the cmdline of the process with the given identifier.
    """
    return (
        subprocess.check_output(f"tr '\\0' ' ' </proc/{pid}/cmdline", shell=True)
        .decode("utf-8")
        .split()
    )


def terminate(pid: int) -> None:
    """
    Terminates the running process with the given pid by sending the signal SIGTERM to it.

    Raises 'ProcessLookupError' if no such process.
    Raises 'PermissionError' if there is no sufficient permissions.
    """
    if int(os.environ.get("GENV_TERMINATE_PROCESSES", "1")) == 0:
        return

    os.kill(pid, signal.SIGTERM)



import argparse
import asyncio
from gettext import gettext as _
import sys

from . import activate
from . import attach
from . import config
from . import deactivate
from . import detach
from . import devices
from . import enforce
from . import envs
from . import home
from . import llm
from . import lock
from . import monitor
from . import remote
from . import shell
from . import status
from . import usage
from . import version


# NOTE(raz): this is needed for modules that their output is being eval() by the
# 'genv' shell function. it is copied almost as-is from argparse.
class _HelpStderrAction(argparse.Action):
    def __init__(
        self,
        option_strings,
        dest=argparse.SUPPRESS,
        default=argparse.SUPPRESS,
        help=None,
    ):
        super(_HelpStderrAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_help(file=sys.stderr)
        parser.exit()


def _is_shell_module(module: str) -> bool:
    """Returns whether a module is a shell module"""

    return module in ["activate", "deactivate"]


def main():
    """
    The genv entrypoint.
    """

    parser = argparse.ArgumentParser(
        description="Query and control Genv on this machine or in a cluster"
    )

    subparsers = parser.add_subparsers(dest="submodule", metavar="SUBCOMMAND")

    for submodule, help, add_arguments in [
        ("activate", "Activate shell environment", activate.add_arguments),
        ("attach", "Attach devices to this environment", attach.add_arguments),
        ("config", "Configure the current environment", config.add_arguments),
        ("deactivate", "Deactivate shell environment", deactivate.add_arguments),
        ("detach", "Attach devices from this environment", detach.add_arguments),
        ("devices", "Query and manage devices", devices.add_arguments),
        ("enforce", "Enforce GPU usage", enforce.add_arguments),
        ("envs", "Query and manage environments", envs.add_arguments),
        (
            "home",
            "Show the home directory of the current environment",
            home.add_arguments,
        ),
        ("llm", "Run and attach to LLMs", llm.add_arguments),
        ("lock", "Lock over-subscribed devices", lock.add_arguments),
        ("monitor", "Monitor using Prometheus and Grafana", monitor.add_arguments),
        ("remote", "Query, manage and monitor remote machines", remote.add_arguments),
        ("shell", "Shell support", shell.add_arguments),
        ("status", "Show status of the current environment", status.add_arguments),
        ("usage", "GPU usage miscellaneous", usage.add_arguments),
        ("version", "Print Genv version", version.add_arguments),
    ]:
        aliases = [f"{submodule}s"] if submodule in ["llm"] else []

        subparser = subparsers.add_parser(
            submodule,
            aliases=aliases,
            help=help,
            add_help=not _is_shell_module(submodule),
        )

        if _is_shell_module(submodule):
            subparser.register("action", "help_stderr", _HelpStderrAction)

            subparser.add_argument(
                "-h",
                "--help",
                action="help_stderr",
                default=argparse.SUPPRESS,
                help=_("show this help message and exit"),
            )

            # this is the process identifier of the shell which is passed by the 'genv'
            # shell function that is installed by "genv shell"
            subparser.add_argument("--shell", type=int, help=argparse.SUPPRESS)

        add_arguments(subparser)

    args = parser.parse_args()

    try:
        if args.submodule is None:
            parser.print_help()
            parser.exit()

        if _is_shell_module(args.submodule) and not args.shell:
            raise RuntimeError(shell.error_msg())

        if args.submodule == "activate":
            activate.run(args.shell, args)
        elif args.submodule == "attach":
            attach.run(args)
        elif args.submodule == "config":
            config.run(args)
        elif args.submodule == "deactivate":
            deactivate.run(args.shell, args)
        elif args.submodule == "detach":
            detach.run(args)
        elif args.submodule == "devices":
            devices.run(args)
        elif args.submodule == "enforce":
            asyncio.run(enforce.run(args))
        elif args.submodule == "envs":
            envs.run(args)
        elif args.submodule == "home":
            home.run(args)
        elif args.submodule in ["llm", "llms"]:
            llm.run(args)
        elif args.submodule == "lock":
            lock.run(args)
        elif args.submodule == "monitor":
            asyncio.run(monitor.run(args))
        elif args.submodule == "remote":
            asyncio.run(remote.run(args))
        elif args.submodule == "shell":
            shell.run(args)
        elif args.submodule == "status":
            status.run(args)
        elif args.submodule == "usage":
            asyncio.run(usage.run(args))
        elif args.submodule == "version":
            version.run(args)
    except RuntimeError as e:
        print(e, file=sys.stderr)
        exit(1)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Iterable, Optional

from .devices import Devices
from .envs import Envs
from .processes import Processes


@dataclass
class Snapshot:
    processes: Processes
    envs: Envs
    devices: Devices

    def filter(
        self,
        deep: bool = True,
        *,
        eid: Optional[str] = None,
        eids: Optional[Iterable[str]] = None,
        username: Optional[str] = None,
    ):
        """
        Returns a new filtered snapshot.

        :param deep: Perform deep filtering
        :param eid: Environment identifier to keep
        :param eids: Environment identifiers to keep
        :param username: Username to keep
        """
        envs = self.envs.filter(deep=deep, eid=eid, eids=eids, username=username)

        return Snapshot(
            processes=self.processes.filter(deep=deep, eids=envs.eids),
            envs=envs,
            devices=self.devices.filter(deep=deep, eids=envs.eids),
        )



import argparse
import getpass
import os

import genv.utils
from genv.entities import Env
import genv.core
import genv.sdk


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv activate" arguments to a parser.
    """

    arguments = parser.add_argument_group("arguments")
    arguments.add_argument("--eid", "--id", help="Environment identifier")

    configure = parser.add_argument_group("configure")
    configure.add_argument("--name", help="Environment name")
    configure.add_argument(
        "--gpu-memory", help="Environment device memory capacity (e.g. 4g)"
    )
    configure.add_argument("--gpus", type=int, help="Environment device count")

    options = parser.add_argument_group("options")
    options.add_argument(
        "--no-prompt",
        action="store_false",
        dest="prompt",
        help="Don't change the prompt",
    )
    options.add_argument(
        "--no-attach",
        action="store_false",
        dest="attach",
        help="Don't attach to devices",
    )
    options.add_argument(
        "-o",
        "--over-subscribe",
        dest="allow_over_subscription",
        action="store_true",
        help="Use unavailable devices if needed",
    )

    load = options.add_mutually_exclusive_group()
    load.add_argument(
        "--no-load",
        action="store_false",
        dest="load",
        help="Don't load configuration from disk (default)",
    )


def run(shell: int, args: argparse.Namespace) -> None:
    """
    Runs the "genv activate" logic.
    """

    if genv.sdk.active():
        raise RuntimeError("Already running in an active environment")

    eid = args.eid or str(shell)

    with genv.utils.global_lock():
        genv.core.envs.activate(
            eid, uid=os.getuid(), username=getpass.getuser(), pid=shell
        )

        # NOTE(raz): we currently override the entire configuration if any
        # configuration field was specified
        if args.name or args.gpu_memory or args.gpus:
            genv.core.envs.configure(
                eid, Env.Config(args.name, args.gpu_memory, args.gpus)
            )

        if args.gpus and args.attach:
            genv.core.devices.attach(
                eid,
                gpus=args.gpus,
                gpu_memory=args.gpu_memory,
                allow_over_subscription=args.allow_over_subscription,
            )

    print(
        f"""
_genv_set_env GENV_SHELL 1
_genv_set_env GENV_ENVIRONMENT_ID {eid}
_genv_replace_env PATH "{os.path.realpath(os.path.join(os.path.dirname(__file__), "../shims"))}:$PATH"

eval "$(command genv shell --reconfigure)"
eval "$(command genv shell --reattach)"
"""
    )

    if args.prompt:
        print('_genv_replace_env PS1 "(genv) ${PS1-}"')



from dataclasses import dataclass
from typing import Iterable


@dataclass
class System:
    """System status."""

    @dataclass
    class Genv:
        """Genv status."""

        installed: bool

    @dataclass
    class Device:
        """Device status."""

        index: int
        utilization: int
        temperature: int
        used_memory: str
        total_memory: str

    genv: Genv
    devices: Iterable[Device]



from typing import Dict

from genv.entities.enforce import Survey


def max_devices_per_user(
    *surveys: Survey, maximum: int, maximum_for_user: Dict[str, int] = {}
) -> None:
    """
    Enforce maximum devices per user.
    """
    usernames = set(
        env.username
        for survey in surveys
        for env in survey.snapshot.envs
        if env.username
    )

    for username in usernames:
        snapshots = [survey.snapshot.filter(username=username) for survey in surveys]

        attached = sum(len(snapshot.devices) for snapshot in snapshots)

        over = attached - maximum_for_user.get(username, maximum)

        if over <= 0:
            continue

        if all(survey.hostname for survey in surveys):
            hosts = len([snapshot for snapshot in snapshots if len(snapshot.envs) > 0])

            print(
                f"User {username} is using {attached} devices on {hosts} hosts which is {over} more than the maximum allowed"
            )
        else:
            print(
                f"User {username} is using {attached} devices which is {over} more than the maximum allowed"
            )

        detached = 0

        for snapshot, survey in zip(snapshots, surveys):
            for device in snapshot.devices:
                if detached == over:
                    break  # already detached enough devices

                survey.detach(device.index, *device.eids)

                detached += 1



import argparse
import json
import sys
from typing import Optional

import genv

# TODO(raz): should this executable be merged to remote.py


async def do_execute() -> None:
    report = json.loads(sys.stdin.read(), cls=genv.JSONDecoder)

    with genv.utils.global_lock():
        genv.enforce.execute(report)


async def do_snapshot(format: str, type: Optional[str]) -> None:
    with genv.utils.global_lock():
        if type is None:
            snapshot = await genv.core.snapshot()
        elif type == "devices":
            snapshot = genv.core.devices.snapshot()
        elif type == "envs":
            snapshot = genv.core.envs.snapshot()
        elif type == "processes":
            snapshot = await genv.core.processes.snapshot()
        else:
            raise ValueError(f"Unsupported snapshot type ({type})")

    if format == "json":
        print(json.dumps(snapshot, cls=genv.JSONEncoder, indent=2))


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv usage" arguments to a parser.
    """

    subparsers = parser.add_subparsers(dest="command", required=True)

    def execute(parser):
        pass

    def snapshot(parser):
        parser.add_argument(
            "--format",
            choices=["json"],
            help="Output format (%(choices)s)",
            default="json",
        )

        parser.add_argument(
            "--type",
            choices=["devices", "envs", "processes"],
            help="Take a snapshot of specific information",
        )

    for command, help in [
        (execute, "Execute the report passed in stdin"),
        (snapshot, "Take a snapshot of GPU usage"),
    ]:
        command(subparsers.add_parser(command.__name__, help=help))


async def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv usage" logic.
    """

    if args.command == "execute":
        await do_execute()
    elif args.command == "snapshot":
        await do_snapshot(args.format, args.type)



from contextlib import contextmanager
from glob import glob
import os
import re
from typing import Iterable, Optional

import genv.utils
import genv.core

from . import env
from .utils import set_temp_env_var


def _visible() -> Iterable[int]:
    """Returns the indices of visible devices"""

    indices = os.environ["CUDA_VISIBLE_DEVICES"]

    return [] if indices == "-1" else [int(index) for index in indices.split(",")]


def _lockable() -> Iterable[int]:
    """Returns the indices of devices that have an existing lock"""

    dir = genv.utils.get_temp_file_path("devices")

    return [
        int(re.sub(r".*?(\d+)\.lock", r"\1", path)) for path in glob(f"{dir}/*.lock")
    ]


def _update_env(indices: Iterable[int]) -> None:
    """Updates the device indices environment variables"""

    set_temp_env_var(
        "CUDA_VISIBLE_DEVICES", ",".join(map(str, indices)) if indices else "-1"
    )  # TODO(raz): support the case when this env var already exists

    # TODO(raz): should we set "CUDA_DEVICE_ORDER" as well?


def attach(
    *,
    index: Optional[int] = None,
    gpus: Optional[int] = None,
    allow_over_subscription: bool = False,
) -> Iterable[int]:
    """Attaches devices to the current environment.

    Attaching to a specific device if an index is specified.
    Attaching to a some devices if count is specified.
    Otherwise, the device count is taken from the environment configuration.

    Does not detach devices if already attached to more devices.
    """

    if index is not None and gpus is not None:
        raise ValueError(
            'Can\'t use both "index" and "count" in genv.sdk.devices.attach()'
        )

    if not env.active():
        raise RuntimeError("Not running in an active environment")

    config = env.configuration()

    kwargs = {}
    if index is not None:
        kwargs["index"] = index
    elif gpus is not None:
        kwargs["gpus"] = gpus
    else:
        kwargs["gpus"] = config.gpus

    with genv.utils.global_lock():
        indices = genv.core.devices.attach(
            env.eid(),
            **kwargs,
            gpu_memory=config.gpu_memory,
            allow_over_subscription=allow_over_subscription,
        )

    _update_env(indices)

    return indices


def attached() -> Iterable[int]:
    """Returns the indices of attached devices.

    Indices are in host namespace even when running in a container.
    Raises RuntimeError if not running in an active environment.
    """

    if not env.active():
        raise RuntimeError("Not running in an active environment")

    # TODO(raz): replacing this logic with an explicit env var "GENV_DEVICE_INDICES"

    if "GENV_SHELL" in os.environ or "GENV_PYTHON" in os.environ:
        return _visible()
    elif "GENV_CONTAINER" in os.environ:
        return _lockable()

    raise RuntimeError("Failed to find attached device indices")


def detach(index: Optional[int] = None) -> Iterable[int]:
    """Detaches the current environment from devices.

    Detaches from a specific device if an index is specified.
    Otherwise, detaches from all devices.

    Raises RuntimeError if not running in an active environment.
    """

    if not env.active():
        raise RuntimeError("Not running in an active environment")

    with genv.utils.global_lock():
        indices = genv.core.devices.detach(env.eid(), index)

    _update_env(indices)

    return indices


def refresh_attached() -> Iterable[int]:
    """Refreshes attached devices.

    Raises RuntimeError if not running in an active environment.
    """

    if not env.active():
        raise RuntimeError("Not running in an active environment")

    with genv.utils.global_lock():
        indices = genv.core.devices.attached(env.eid())

    _update_env(indices)

    return indices


@contextmanager
def lock() -> None:
    """Obtains exclusive access to the attached devices.

    Does nothing if not running in an active environment or not attached to devices.
    """

    if not env.active():
        yield
    else:
        indices = attached()

        with genv.core.devices.lock(indices):
            yield



from ..entities import Env
from .devices import attach, attached, detach, lock, refresh_attached
from .env import (
    active,
    activate,
    configure,
    configuration,
    eid,
    home,
    load_configuration,
    refresh_configuration,
    save_configuration,
)



import asyncio
from asyncio.subprocess import Process
from typing import Dict, Optional

from .runner import Runner as Base


class Runner(Base):
    hostname: str
    username: Optional[str]
    timeout: Optional[int]

    def __init__(
        self,
        hostname: str,
        username: Optional[str] = None,
        timeout: Optional[int] = None,
        process_env: Optional[Dict[str, str]] = None,
    ):
        super().__init__(process_env)
        self.hostname = hostname
        self.username = username
        self.timeout = timeout

    async def _open_process(self, *args: str, stdin_fd: int, sudo: bool) -> Process:
        ssh_parameters = []

        if self.username:
            ssh_parameters.extend(["-l", self.username])

        if self.timeout is not None:
            ssh_parameters.extend(["-o", f"ConnectTimeout={self.timeout}"])

        remote_command = " ".join(args)

        if self._process_env:
            remote_command = (
                "env "
                + " ".join(
                    [f"{name}={value}" for name, value in self._process_env.items()]
                )
                + " "
                + remote_command
            )

        if sudo:
            remote_command = f"sudo {remote_command}"

        return await asyncio.create_subprocess_exec(
            "ssh",
            *ssh_parameters,
            self.hostname,
            remote_command,
            stdin=stdin_fd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )



import sys
from typing import Dict

import genv.utils
from genv.entities import Envs, Processes, Report
import genv.core.devices


def _terminate(processes: Processes) -> None:
    for process in processes:
        try:
            print(
                f"Terminating process {process.pid} from environment {process.eid or 'N/A'} that is running on GPU(s) {','.join([str(index) for index in process.indices])}"
            )

            genv.utils.terminate(process.pid)
        except PermissionError:
            print(
                f"[ERROR] Not enough permissions to terminate process {process.pid}",
                file=sys.stderr,
            )
        except ProcessLookupError:
            print(f"[DEBUG] Process {process.pid} already terminated", file=sys.stderr)


def _detach(envs: Dict[int, Envs]) -> None:
    for index, envs in envs.items():
        for env in envs:
            print(
                f"Detaching environment {env.eid} of user {env.username or 'N/A'} from device {index}"
            )

            genv.core.devices.detach(env.eid, index)


def execute(report: Report) -> None:
    """
    Terminates processes and detaches environments from devices according to the given report.
    """
    _terminate(report.terminate)
    _detach(report.detach)



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, Optional, Set

from ..core import Snapshot

from . import Report


@dataclass
class Survey:
    snapshot: Snapshot
    hostname: Optional[str] = None

    _pids: Set[int] = field(default_factory=set, init=False)
    _eids: Dict[int, Set[str]] = field(
        default_factory=lambda: defaultdict(set), init=False
    )

    def terminate(self, *pids: int) -> None:
        self._pids.update(pids)

    def detach(self, index: int, *eids: str) -> None:
        self._eids[index].update(eids)

        self.terminate(
            *(
                process.pid
                for eid in eids
                for process in self.snapshot.processes.filter(eid=eid, index=index)
            )
        )

    @property
    def report(self) -> Report:
        return Report(
            terminate=self.snapshot.processes.filter(deep=False, pids=self._pids),
            detach={
                index: self.snapshot.envs.filter(deep=False, eids=eids)
                for index, eids in self._eids.items()
            },
        )



from typing import Iterable, Optional

from genv.entities import Snapshot, System

from .metric import Metric
from .spec import Spec
from .type import Type


class Collection:
    """
    A metric collection.
    """

    def __init__(self, specs: Iterable[Spec]) -> None:
        self._metrics = [
            Metric(
                spec.name,
                spec.documentation,
                spec.labelnames,
                spec=spec,
            )
            for spec in specs
        ]

    def __iter__(self):
        return self._metrics.__iter__()

    def __getitem__(self, name: str) -> Metric:
        return next(metric for metric in self if metric.name == name)

    def _find(self, type: Type) -> Iterable[Metric]:
        """
        Returns all metrics of the given type.
        """
        return [metric for metric in self if metric.spec.type == type]

    def cleanup(
        self,
        system: Optional[System],
        snapshot: Optional[Snapshot],
        *,
        header: Optional[str] = None
    ) -> None:
        """
        Cleans up metric label values.

        :param header: First label value to filter; optional.
        """
        for metric in self:
            for labelvalues in metric.label_sets():
                if header and labelvalues[0] != header:
                    continue

                if (
                    system
                    and not system.genv.installed
                    and metric.spec.type not in [Type.General, Type.Device]
                ):
                    metric.remove(*labelvalues)

                if snapshot and metric.spec.filter:
                    if not metric.spec.filter(
                        labelvalues[1:] if header else labelvalues, snapshot
                    ):
                        metric.remove(*labelvalues)

    def update(
        self, system: Optional[System], snapshot: Optional[Snapshot], labels: dict = {}
    ) -> None:
        """
        Updates metric values.
        """
        if system:
            self._general(system, labels)
            self._device(system, labels)

        if snapshot:
            self._system(snapshot, labels)
            self._env(snapshot, labels)
            self._process(snapshot, labels)
            self._user(snapshot, labels)

    def _general(self, system: System, labels: dict) -> None:
        """Updates general metrics."""

        self["genv_is_installed"].labels(**labels).set(system.genv.installed)

    def _device(self, system: System, labels: dict) -> None:
        """
        Updates per-device metrics.
        """
        for device in system.devices:
            for metric in self._find(Type.Device):
                if not metric.spec.convert:
                    continue

                metric.labels(index=device.index, **labels).set(
                    metric.spec.convert(device)
                )

    def _system(self, snapshot: Snapshot, labels: dict) -> None:
        """
        Updates system-wide metrics.
        """
        for metric in self._find(Type.System):
            if not metric.spec.convert:
                continue

            metric.labels(**labels).set(metric.spec.convert(snapshot))

    def _env(self, snapshot: Snapshot, labels: dict) -> None:
        """
        Updates per-environment metrics.
        """
        for env in snapshot.envs:
            env_snapshot = snapshot.filter(eid=env.eid)

            for metric in self._find(Type.Environment):
                if not metric.spec.convert:
                    continue

                metric.labels(eid=env.eid, **labels).set(
                    metric.spec.convert(env_snapshot)
                )

    def _process(self, snapshot: Snapshot, labels: dict) -> None:
        """
        Updates per-process metrics.
        """
        for process in snapshot.processes:
            pid = process.pid
            eid = process.eid or ""

            self["genv_process_devices_total"].labels(pid=pid, eid=eid, **labels).set(
                len(process.indices)
            )

            for usage in process.used_gpu_memory:
                self["genv_process_used_gpu_memory_bytes"].labels(
                    pid=pid, eid=eid, device=usage.index, **labels
                ).set(usage.bytes)

    def _user(self, snapshot: Snapshot, labels: dict) -> None:
        """
        Updates per-user metrics.
        """
        for username in snapshot.envs.usernames:
            user_snapshot = snapshot.filter(username=username)

            for metric in self._find(Type.User):
                if not metric.spec.convert:
                    continue

                metric.labels(username=username, **labels).set(
                    metric.spec.convert(user_snapshot)
                )



from dataclasses import dataclass
from datetime import datetime
from typing import Callable, Iterable, Optional, Union

from genv.utils import DATETIME_FMT, memory_to_bytes


@dataclass
class Device:
    @dataclass
    class Attachement:
        eid: str
        gpu_memory: Optional[str]
        time: str

    index: int
    total_memory: str
    attachments: Iterable[Attachement]

    @property
    def eids(self) -> Iterable[str]:
        return [attachment.eid for attachment in self.attachments]

    @property
    def attached(self) -> bool:
        return len(self.eids) > 0

    @property
    def detached(self) -> bool:
        return len(self.eids) == 0

    @property
    def total_memory_bytes(self) -> int:
        """
        Returns the device total memory in bytes.
        """
        return memory_to_bytes(self.total_memory)

    @property
    def available_memory_bytes(self) -> int:
        """
        Returns the device available memory in bytes.
        """
        available_bytes = memory_to_bytes(self.total_memory)

        for attachment in self.attachments:
            available_bytes -= memory_to_bytes(
                attachment.gpu_memory or self.total_memory
            )

        return max(available_bytes, 0)

    def available(self, gpu_memory: Optional[str]) -> bool:
        """
        Returns is the device is available with respect to the given memory specification.
        If memory is specified, checks if this amount is available.
        Otherwise, checks if the device is detached.
        """
        if gpu_memory is None:
            return self.detached

        return self.available_memory_bytes >= memory_to_bytes(gpu_memory)

    def filter(self, *, eids: Iterable[str]):
        """
        Returns a new device with only the given environment identifiers.
        """
        return Device(
            self.index,
            self.total_memory,
            [attachment for attachment in self.attachments if attachment.eid in eids],
        )

    def attach(self, eid: str, gpu_memory: Optional[str], time: str) -> None:
        """
        Attaches an environment.
        """
        self.attachments.append(Device.Attachement(eid, gpu_memory, time))

    def detach(self, eid: str) -> None:
        """
        Detaches an environment.
        """
        for index, attachment in enumerate(self.attachments):
            if attachment.eid == eid:
                del self.attachments[index]


@dataclass
class Devices:
    """
    A collection of devices.
    """

    devices: Iterable[Device]

    @property
    def indices(self) -> Iterable[int]:
        return [device.index for device in self.devices]

    def __iter__(self):
        return self.devices.__iter__()

    def __len__(self):
        return self.devices.__len__()

    def __getitem__(self, index: int) -> Device:
        return next(device for device in self.devices if device.index == index)

    def __contains__(self, index: int) -> bool:
        return index in self.indices

    def filter(
        self,
        deep: bool = True,
        *,
        indices: Optional[Iterable[int]] = None,
        not_indices: Optional[Iterable[int]] = None,
        eid: Optional[str] = None,
        eids: Optional[Iterable[str]] = None,
        attached: Optional[bool] = None,
        function: Optional[Callable[[Device], bool]] = None,
    ):
        """
        Returns a new filtered collection.

        :param deep: Perform deep filtering
        :param indices: Device indices to keep
        :param not_indices: Device indices to remove
        :param eid: Environment identifier to keep
        :param eids: Environment identifiers to keep
        :param attached: Keep only devices with environments attached or not
        :param function: Keep devices on which the lambda returns True
        """
        if eids:
            eids = set(eids)

        if eid:
            if not eids:
                eids = set()

            eids.add(eid)

        devices = self.devices

        if indices is not None:
            devices = [device for device in devices if device.index in indices]

        if not_indices is not None:
            devices = [device for device in devices if device.index not in not_indices]

        if eids is not None:
            if deep:
                devices = [device.filter(eids=eids) for device in devices]

            devices = [
                device
                for device in devices
                if any((eid in eids) for eid in device.eids)
            ]

        if attached is not None:
            if attached:
                devices = [device for device in devices if device.attached]
            else:
                devices = [device for device in devices if device.detached]

        if function is not None:
            devices = [device for device in devices if function(device)]

        return Devices(devices)

    def attach(
        self, eid: str, indices: Union[Iterable[int], int], gpu_memory: Optional[str]
    ) -> None:
        """
        Attaches devices to an environment.
        """

        if isinstance(indices, int):
            indices = [indices]

        time = datetime.now().strftime(DATETIME_FMT)

        for index in indices:
            self.devices[index].attach(eid, gpu_memory, time)

    def detach(self, eid: str, index: Optional[int] = None) -> None:
        """Detaches an environment"""

        if index is not None:
            self.devices[index].detach(eid)
        else:
            for device in self.devices:
                device.detach(eid)

    def cleanup(
        self,
        *,
        poll_eid: Optional[Callable[[str], bool]] = None,
    ):
        """
        Cleans up the collection in place.
        """
        for device in self.devices:
            for eid in device.eids:
                if (poll_eid is not None) and (not poll_eid(eid)):
                    device.detach(eid)

    def find_available_devices(
        self, count: int, gpu_memory: Optional[str], allow_over_subscription: bool
    ) -> Iterable[int]:
        """
        Finds available devices with respect to the optionally specified memory request.
        When thare are not enough devices available, non available devices are also used if
        over-subscription is allowed. Otherwise, RuntimeError is raised.

        :param count: Amount of devices to find
        :param gpu_memory: Optional memory request
        :param allow_over_subscription: Use non available devices if needed

        :return: A list of device indices
        """
        available_devices = self.filter(
            function=lambda device: device.available(gpu_memory)
        )

        indices = available_devices.indices

        if allow_over_subscription and len(indices) < count:
            not_available_devices = self.filter(not_indices=available_devices.indices)

            # we sort in order to use the least populated devices first
            indices += sorted(
                not_available_devices.indices,
                key=lambda index: len(self[index].eids),
            )

        if len(indices) < count:
            raise RuntimeError("No available devices")

        return indices[:count]



import argparse
import sys
from typing import Callable, Iterable, Optional

import genv

QUERIES = {
    "index": lambda device: device.index,
    "eids": lambda device: " ".join(device.eids),
    "total_memory": lambda device: device.total_memory,
    "attachments": lambda device: " ".join(
        f"{attachment.eid}+{attachment.gpu_memory or ''}+{attachment.time.replace(' ', '_')}"
        for attachment in device.attachments
    ),
}


def cleanup(
    snapshot: genv.Devices,
    eid: Optional[str] = None,
    count: Optional[int] = None,
) -> None:
    """
    Detaches devices from environments.

    :param snapshot:
    :param eid: If passed, detaches only from this environment. Otherwise, detaches all deactivated environments.
    :param count: Stops when this amount of devices are detached.
    :return: None
    """
    pred: Callable[[str], bool]

    if eid is not None:
        pred = lambda _: _ == eid
    else:
        envs = genv.core.envs.snapshot()
        pred = lambda eid: eid not in envs.eids

    for index in sorted(snapshot.indices, reverse=True):
        if count == 0:
            break

        device = snapshot[index]

        if device.detached:
            continue

        for candidate in device.eids:
            if not pred(candidate):
                continue

            device.detach(candidate)

        # TODO(raz): we have a bug here when an environment is attached to multiple devices
        # along with other environments.
        # we don't properly check if the detachment took place as we check if the device
        # is detached entirely.
        # therefore, 'count' is not respected and all devices are detached.

        if device.detached and count is not None:
            count -= 1


def do_attach(
    snapshot: genv.Devices,
    eid: str,
    count: Optional[int],
    index: Optional[int],
    allow_over_subscription: bool,
) -> None:
    """
    Attaches devices to an environment and prints the attached device indices.
    Either ensures that a specified count is attached to the environment
    or explicitly tries to attach a device with a given index.

    :param snapshot:
    :param eid: Environment identifier
    :param count: Amount of devices to attach
    :param index: Device index to attach
    :return: None
    """
    # TODO(raz): merge this method into genv.core.devices.attach() entirely

    envs = genv.core.envs.snapshot()

    if count is None and index is None:
        # use the environment device count configuration if set
        count = envs[eid].config.gpus if eid in envs else None

    gpu_memory = envs[eid].config.gpu_memory if eid in envs else None
    attached_devices = snapshot.filter(eid=eid)

    def ensure_count():
        """
        Attach or detach devices from the environment according to the number of devices
        set in the environment configuration.
        """
        current = len(attached_devices)

        if current < count:
            if count > len(snapshot):
                raise RuntimeError(
                    f"Requested more devices ({count}) than total available ({len(snapshot)})"
                )

            not_attached_devices = snapshot.filter(not_indices=attached_devices.indices)

            indices = not_attached_devices.find_available_devices(
                count - current, gpu_memory, allow_over_subscription
            )

            snapshot.attach(eid, indices, gpu_memory)
        elif current > count:
            cleanup(snapshot, eid, current - count)

    def attach_by_index():
        """
        Attach a specific device to the environment.
        """
        if index not in attached_devices.indices:
            if not allow_over_subscription and not snapshot[index].available(
                gpu_memory
            ):
                raise RuntimeError(f"Device {index} is not available")

            snapshot.attach(eid, index, gpu_memory)

    # TODO(raz): should this try/except be moved to main()?
    try:
        if count is not None:
            ensure_count()
        elif index is not None:
            attach_by_index()
    except RuntimeError as e:
        print(e, file=sys.stderr)
        exit(1)

    do_find(snapshot, eid)  # print the up-to-date device indices


def do_detach(
    snapshot: genv.Devices, eid: str, index: Optional[int], quiet: bool
) -> None:
    """
    Detaches devices from an environment. After detaching, print environment-devices current relationship.
    Either the device to detach is given, or a general cleanup is called on the environment.

    :param snapshot:
    :param eid: Environment identifier
    :param index: An index of the specific device to detach.
                If no index is given a general cleanup is called on the environment.
    :return: None
    """
    if index is not None:
        if index in snapshot.indices:
            snapshot[index].detach(eid)
    else:
        cleanup(snapshot, eid)

    if not quiet:
        do_find(snapshot, eid)  # print the up-to-date device indices


def do_find(snapshot: genv.Devices, eid: str) -> None:
    """
    Prints the indices of devices attached to the given environment.
    """

    print(",".join(str(index) for index in snapshot.filter(eid=eid).indices))


def do_ps(snapshot: genv.Devices, format: str, header: bool, timestamp: bool) -> None:
    """
    Prints information in a human-readable fashion to the standard output.

    :param snapshot:
    :param format: Which format should be used to print the devices data - "csv" or "tui".
    :param header: a flag instructing if we should print a header as part of the output.
    :param timestamp: a flag instructing if we should print attachment time as is or relative duration (now - creation).
    :return: None
    """
    if header:
        if format == "csv":
            print("ID,ENV ID,ENV NAME,ATTACHED")
        elif format == "tui":
            print("ID      ENV ID      ENV NAME        ATTACHED")

    def print_line(index, eid, name: Optional[str], attached):
        name = name or ""

        if format == "csv":
            print(f"{index},{eid},{name},{attached}")
        elif format == "tui":
            print(f"{index:<8}{eid:<12}{name:<16}{attached}")

    envs = genv.core.envs.snapshot()

    for device in snapshot:
        if device.detached:
            print_line(device.index, "", "", "")
        else:
            for attachment in device.attachments:
                index = device.index
                eid = attachment.eid
                name = envs[eid].config.name if eid in envs else None

                attached = (
                    attachment.time
                    if timestamp
                    else genv.utils.time_since(attachment.time)
                )

                print_line(index, eid, name, attached)


def do_query(snapshot: genv.Devices, queries: Iterable[str]) -> None:
    """
    Queries devices.
    """
    for device in snapshot:

        def query(name: str) -> str:
            result = QUERIES[name](device)

            return "" if result is None else str(result)

        print(",".join(query(name) for name in queries))


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv devices" arguments to a parser.
    """

    parser.add_argument(
        "--no-cleanup",
        dest="cleanup",
        action="store_false",
        help="Do not perform clean up",
    )

    parser.add_argument("--reset", action="store_true", help="Reset previous state")

    subparsers = parser.add_subparsers(dest="command")

    def attach(parser):
        parser.add_argument("--eid", required=True, help="Environment identifier")
        parser.add_argument(
            "--allow-over-subscription",
            action="store_true",
            help="Use unavailable devices if needed",
        )
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            # TODO(raz): remove this argument entirely.
            # either attach an environment to devices based on its configuration or by specific index
            "--count",
            type=int,
            help="Total devices count to be attached",
        )
        group.add_argument(
            "--index", type=int, help="Device index to attach specifically"
        )

    def detach(parser):
        parser.add_argument("--eid", required=True, help="Environment identifier")
        parser.add_argument("--index", type=int, help="Device index to dettach")
        parser.add_argument(
            "--quiet",
            action="store_true",
            help="Don't print updated device indexes",
        )

    def find(parser):
        parser.add_argument("--eid", required=True, help="Environment identifier")

    def ps(parser):
        parser.add_argument(
            "--no-header",
            dest="header",
            action="store_false",
            help="Do not print column headers",
        )
        parser.add_argument(
            "--timestamp", action="store_true", help="Print a non-prettified timestamp"
        )
        parser.add_argument(
            "--format",
            choices=["csv", "tui"],
            help="Output format; CSV or TUI (Text-based user interface)",
            default="tui",
        )

    def query(parser):
        parser.add_argument(
            "--query",
            "--queries",
            dest="queries",
            nargs="+",
            choices=QUERIES.keys(),
            required=True,
        )

    for command, help in [
        (attach, "Attach devices to an environment"),
        (detach, "Detach devices from an environment"),
        (find, "Find the devices attached to an enviornment"),
        (ps, "Print information about devices"),
        (query, "Query devices"),
    ]:
        command(subparsers.add_parser(command.__name__, help=help))


def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv devices" logic.
    """

    with genv.utils.global_lock():
        with genv.core.devices.State(args.cleanup, args.reset) as devices:
            if args.command == "attach":
                do_attach(
                    devices,
                    args.eid,
                    args.count,
                    args.index,
                    args.allow_over_subscription,
                )
            elif args.command == "detach":
                do_detach(devices, args.eid, args.index, args.quiet)
            elif args.command == "find":
                do_find(devices, args.eid)
            elif args.command == "ps":
                do_ps(devices, args.format, args.header, args.timestamp)
            elif args.command == "query":
                do_query(devices, args.queries)
            else:
                do_ps(devices, format="tui", header=True, timestamp=False)



import argparse
import asyncio

import genv


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv enforce" arguments to a parser.
    """

    parser.add_argument(
        "--interval",
        type=int,
        default=10,
        help="interval in seconds between enforcement cycles; 0 means run once (default: %(default)s)",
    )

    enforcements = parser.add_argument_group("enforcements")

    def add_enforcement(
        name: str, *, dest: str, help: str, default: bool = True
    ) -> None:
        """
        Adds a mutual exclusive group for an enforcement flag with true and false actions.
        """
        group = enforcements.add_mutually_exclusive_group()

        group.add_argument(
            f"--{name}",
            action="store_true",
            default=default,
            help=help + " (default: %(default)s)",
            dest=dest,
        )

        group.add_argument(
            f"--no-{name}",
            action="store_false",
            dest=dest,
        )

    add_enforcement(
        "non-env-processes",
        dest="non_env_processes",
        help="terminate processes that are not running in a GPU environment",
        default=False,
    )

    add_enforcement(
        "env-devices",
        dest="env_devices",
        help="enforce environment attached devices",
        default=True,
    )

    add_enforcement(
        "env-memory",
        dest="env_memory",
        help="enforce environment memory capacity if set",
        default=True,
    )

    enforcements.add_argument(
        "--max-devices-per-user",
        type=int,
        help="maximum allowed attached devices for each user",
    )

    def max_devices_for_user(value: str):
        try:
            username, maximum = value.split("=")

            return username, int(maximum)
        except (ValueError, SyntaxError):
            raise argparse.ArgumentTypeError(f"not a valid spec: {value}")

    enforcements.add_argument(
        "--max-devices-for-user",
        nargs="+",
        help="per-user specification of maximum allowed attached devices",
        metavar="username=maximum",
        type=max_devices_for_user,
    )


async def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv enforce" logic.
    """

    while True:
        with genv.utils.global_lock():
            snapshot = await genv.core.snapshot()

        survey = genv.entities.enforce.Survey(snapshot)

        if args.non_env_processes:
            genv.enforce.rules.non_env_processes(survey)

        if args.env_devices:
            genv.enforce.rules.env_devices(survey)

        if args.env_memory:
            genv.enforce.rules.env_memory(survey)

        if args.max_devices_per_user is not None:
            genv.enforce.rules.max_devices_per_user(
                survey,
                maximum=args.max_devices_per_user,
                maximum_for_user=(
                    dict(args.max_devices_for_user) if args.max_devices_for_user else {}
                ),
            )

        with genv.utils.global_lock():
            genv.enforce.execute(survey.report)

        if args.interval == 0:
            break

        await asyncio.sleep(args.interval)



from contextlib import contextmanager
from datetime import datetime
import json
import os
from typing import Any, Callable, Optional, Type, TypeVar, Union

from .os_ import Umask, access_lock

T = TypeVar("T")

DATETIME_FMT = "%d/%m/%Y %H:%M:%S"
MEMORY_TO_BYTES_MULTIPLIERS_DICT = {
    "b": 1,
    "k": 1000,
    "m": 1000 * 1000,
    "g": 1000 * 1000 * 1000,
    "ki": 1024,
    "mi": 1024 * 1024,
    "gi": 1024 * 1024 * 1024,
}


def get_temp_file_path(filename: str) -> str:
    """
    Returns the path of a file with the provided name in the Genv temporary directory.
    """
    return os.path.join(os.environ.get("GENV_TMPDIR", "/var/tmp/genv"), filename)


def load_state(
    path: str,
    *,
    creator: Callable[[], T],
    cleaner: Callable[[T], None],
    converter: Optional[Callable[[Any], T]] = None,
    json_decoder: Optional[Type[json.JSONDecoder]] = None,
    cleanup: bool = True,
    reset: bool = False,
) -> T:
    """
    Loads a state file.
    """
    if os.path.exists(path) and not reset:
        with open(path) as f:
            o = json.load(f, cls=json_decoder)

        if converter:
            o = converter(o)

        if cleanup:
            cleaner(o)
    else:
        o = creator()

    return o


def save_state(
    o: Any,
    path: str,
    *,
    json_encoder: Optional[Type[json.JSONEncoder]] = None,
):
    """
    Saves a state file.
    """
    with Umask(0):
        with open(
            path, "w", opener=lambda path, flags: os.open(path, flags, 0o666)
        ) as f:
            json.dump(o, f, cls=json_encoder, indent=2)


def memory_to_bytes(cap: str) -> int:
    """
    Convert memory string to an integer value in bytes.
    """
    for unit, multiplier in MEMORY_TO_BYTES_MULTIPLIERS_DICT.items():
        if cap.endswith(unit):
            return int(cap.replace(unit, "")) * multiplier

    return int(cap)  # the value is already in bytes if no unit was specified


# TODO(raz): support detecting the most suitable units automatically
def bytes_to_memory(bytes: int, unit: str, suffix: bool = True) -> str:
    """
    Convert bytes to a memory string.
    """
    s = f"{bytes // MEMORY_TO_BYTES_MULTIPLIERS_DICT[unit]}"

    if suffix:
        s = f"{s}{unit}"

    return s


def memory_to_memory(memory: str, unit: str, suffix: bool = True) -> str:
    """
    Convert memory string to a memory string in another unit.
    """
    return bytes_to_memory(bytes=memory_to_bytes(memory), unit=unit, suffix=suffix)


def time_since(dt: Union[str, datetime]) -> str:
    """
    This function returns a human readable string describing the amount of time passed since 'dt'
    :param dt: The base time to calculate from. Can be either string or datetime
    :return: a human readable string describing the amount of time passed since 'dt'
    """
    if isinstance(dt, str):
        dt = datetime.strptime(dt, DATETIME_FMT)

    value = int((datetime.now() - dt).total_seconds())
    unit = "second"

    for amount, next_units in [
        (60, "minute"),
        (60, "hour"),
        (24, "day"),
        (7, "week"),
    ]:
        if value < amount:
            break

        value, _ = divmod(value, amount)
        unit = next_units

    if value > 1:
        unit = f"{unit}s"

    return f"{value} {unit} ago"


@contextmanager
def global_lock() -> None:
    """
    Locks the global lock.
    """
    with access_lock(get_temp_file_path("genv.lock")):
        yield



import argparse
from dataclasses import dataclass
from typing import Optional, Type

import genv.sdk
from genv.sdk import Env


@dataclass
class Field:
    name: str
    type: Optional[Type] = None

    @property
    def human_readable_name(self) -> str:
        """Converts the field name to a human readable one"""

        return self.name.replace("_", "-")


FIELDS = [
    Field("gpu_memory"),
    Field("gpus", type=int),
    Field("name"),
]


def print_field(config: Env.Config, field: Field, *, prefix: bool = False) -> None:
    """Prints a field"""

    value = getattr(config, field.name)

    if value is not None:
        print(f"{field.human_readable_name}: {value}" if prefix else value)


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv config" arguments to a parser.
    """

    parser.add_argument("-q", "--quiet", "--refresh", action="store_true")
    parser.add_argument("--clear", action="store_true")
    parser.add_argument(
        "--load", action="store_true", help="Load configuration from disk"
    )
    parser.add_argument(
        "--save", action="store_true", help="Save updated configuration to disk"
    )

    subparsers = parser.add_subparsers(dest="field")

    for field in FIELDS:
        subparser = subparsers.add_parser(field.human_readable_name)
        subparser.add_argument(field.name, type=field.type, nargs="?")


def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv config" logic.
    """

    if args.load:
        config = genv.sdk.load_configuration()
    else:
        config = genv.sdk.refresh_configuration()

    if args.field:
        field = next(
            field for field in FIELDS if field.human_readable_name == args.field
        )

        value = getattr(args, field.name)

        if args.clear:
            setattr(config, field.name, None)
        elif value is not None:
            setattr(config, field.name, value)
        elif not args.quiet:
            print_field(config, field)
    else:
        if args.clear:
            config = genv.sdk.Env.Config()
        elif not args.quiet:
            for field in FIELDS:
                print_field(config, field, prefix=True)

    genv.sdk.configure(config)

    if args.save:
        genv.sdk.save_configuration()



from typing import Iterable, Tuple

from genv.entities import System

from ..utils import run, Host, Config, Command


async def system(config: Config) -> Tuple[Iterable[Host], Iterable[System]]:
    """Returns information about remote systems"""

    GENV_COMMAND = "command -v genv 2> /dev/null || echo none"
    DEVICES_COMMAND = "nvidia-smi --query-gpu=index,utilization.gpu,temperature.gpu,memory.used,memory.total --format=csv,noheader,nounits"

    command = Command(
        ["bash", "-c", f'"({GENV_COMMAND}) && ({DEVICES_COMMAND})"'], shell=True
    )

    hosts, stdouts = await run(config, command)

    def parse_stdout(stdout: str) -> System:
        genv_info, device_infos = stdout.split("\n", 1)

        def parse_genv_info(info: str) -> System.Genv:
            return System.Genv(installed=info != "none")

        def parse_device_info(line: str) -> System.Device:
            index, utilization, temperature, used_memory, total_memory = line.split(
                ", "
            )

            return System.Device(
                index=int(index),
                utilization=int(utilization),
                temperature=int(temperature),
                used_memory=f"{used_memory}mi",
                total_memory=f"{total_memory}mi",
            )

        return System(
            genv=parse_genv_info(genv_info),
            devices=[parse_device_info(line) for line in device_infos.splitlines()],
        )

    systems = [parse_stdout(stdout) for stdout in stdouts]

    return hosts, systems



from dataclasses import dataclass
from typing import Dict

from ..core import Envs, Processes


@dataclass
class Report:
    terminate: Processes
    detach: Dict[int, Envs]

    def __bool__(self) -> bool:
        return len(self.terminate) > 0 or len(self.detach) > 0



import sys
from typing import Iterable, TextIO


def reprint(
    hostnames: Iterable[str], outputs: Iterable[str], file: TextIO = sys.stdout
) -> None:
    """
    Reprints outputs from multiple hosts with the hostname as a prefix.

    :param hostnames: Hostnames or IP addresses
    :param outputs: Output for every host
    :param file: Output to print to
    """
    for hostname, output in zip(hostnames, outputs):
        for line in output.splitlines():
            print(f"[{hostname}] {line}", file=file)



from typing import Iterable, Tuple

from genv.entities import Devices

from .snapshot import exec, Config, Host


async def snapshot(config: Config) -> Tuple[Iterable[Host], Iterable[Devices]]:
    """
    Takes device snapshots on multiple hosts.

    :return: Returns the hosts that succeeded and their snapshots
    """
    return await exec(config, type="devices", sudo=False)



import argparse

import genv.sdk


def do_init() -> None:
    """Prints the init shell script"""

    print(
        """\
_genv_append_to_env()
{
  # based on https://unix.stackexchange.com/a/415028
  export $1="${!1:+${!1}:}$2"
}

_genv_backup_env()
{
  if [ -n "${!1}" ]; then
    export GENV_BACKUP_ENV_$1="${!1}"
    _genv_append_to_env GENV_BACKUP_ENVS $1
  fi
}

_genv_set_env()
{
  export $1="$2"
  _genv_append_to_env GENV_ENVS $1
}

_genv_replace_env()
{
  _genv_backup_env $1
  _genv_set_env $1 "$2"
}

_genv_unset_env()
{
  unset $1
  # TODO(raz): remove from 'GENV_ENVS'
}

_genv_unset_envs()
{
  IFS=: read -a names <<< "$GENV_ENVS"
  unset GENV_ENVS

  for name in "${names[@]}"
  do
    unset $name
  done
}

_genv_restore_env()
{
  backup="GENV_BACKUP_ENV_$1"
  if [ -n "${!backup}" ]; then
    export $1="${!backup}"
  fi
  unset $backup
}

_genv_restore_envs()
{
  IFS=: read -a names <<< "$GENV_BACKUP_ENVS"
  unset GENV_BACKUP_ENVS

  for name in "${names[@]}"
  do
    _genv_restore_env $name
  done
}

genv()
{
  local command="${1:-}"
  if [ "$#" -gt 0 ]; then
    shift
  fi

  case "$command" in
  activate|deactivate)
    eval "$(command genv $command --shell $$ $@)"
    ;;
  config)
    command genv config $@

    if [ "$?" -eq 0 ]; then
      eval "$(command genv shell --reconfigure)"
    fi
    ;;
  attach|detach)
    command genv $command $@

    if [ "$?" -eq 0 ]; then
      eval "$(command genv shell --reattach)"
    fi
    ;;
  shell)
    if [ "$#" -eq 0 ]; then
      command genv shell --ok
    else
      command genv shell $@
    fi
    ;;
  *)
    command genv $command $@
    ;;
  esac
}

if [ "$GENV_PREACTIVATE" = "1" ]; then
  command="genv activate"

  if [ "$GENV_PREACTIVATE_PROMPT" = "0" ]; then
    command="$command --no-prompt"
  fi

  if [ -n "$GENV_PREACTIVATE_GPUS" ]; then
    command="$command --gpus $GENV_PREACTIVATE_GPUS"
  fi

  if [ -n "$GENV_PREACTIVATE_ENVIRONMENT_NAME" ]; then
    command="$command --name $GENV_PREACTIVATE_ENVIRONMENT_NAME"
  fi

  eval "$command"
fi
"""
    )


def do_deinit() -> None:
    """Prints the deinit shell script"""

    print(
        """\
unset -f genv
unset -f _genv_restore_envs
unset -f _genv_restore_env
unset -f _genv_unset_envs
unset -f _genv_unset_env
unset -f _genv_replace_env
unset -f _genv_set_env
unset -f _genv_backup_env
unset -f _genv_append_to_env
"""
    )


def error_msg() -> str:
    """Returns an error message"""

    return """\
Your shell is not properly initialized at the moment.
Run the following command to initialize it.
You should also add it to your ~/.bashrc or any equivalent file.

    eval "$(genv shell --init)"
"""


def do_error() -> None:
    """Prints an error message"""

    print(error_msg())


def do_ok() -> None:
    """Prints an ok message"""

    print(
        """\
Your shell is initialized properly and you are all set.
Run the following command to check the status of your environment:

    genv status

If you are not sure how to continue from here, check out the quick start tutorial at https://docs.genv.dev/overview/quickstart.html.
"""
    )


def do_reattach() -> None:
    """Refreshes the shell indices environment variables"""

    indices = genv.sdk.refresh_attached()

    print(
        f"""\
_genv_replace_env CUDA_VISIBLE_DEVICES {",".join(map(str, indices)) if indices else "-1"}
"""
    )


def do_reconfigure() -> None:
    """Refreshes the shell configuration environment variables"""

    config = genv.sdk.refresh_configuration()

    for name, value in [
        ("GENV_ENVIRONMENT_NAME", config.name),
        ("GENV_GPU_MEMORY", config.gpu_memory),
        ("GENV_GPUS", config.gpus),
    ]:
        if value is not None:
            print(f"_genv_set_env {name} {value}")
        else:
            print(f"_genv_unset_env {name}")


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv shell" arguments to a parser.
    """

    parser.add_argument(
        "--init",
        action="store_const",
        dest="action",
        const="init",
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        "--deinit",
        action="store_const",
        dest="action",
        const="deinit",
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        "--ok",
        action="store_const",
        dest="action",
        const="ok",
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        "--reattach",
        action="store_const",
        dest="action",
        const="reattach",
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        "--reconfigure",
        action="store_const",
        dest="action",
        const="reconfigure",
        help=argparse.SUPPRESS,
    )


def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv shell" logic.
    """

    if args.action == "init":
        do_init()
    elif args.action == "deinit":
        do_deinit()
    elif args.action == "ok":
        do_ok()
    elif args.action == "reattach":
        do_reattach()
    elif args.action == "reconfigure":
        do_reconfigure()
    else:
        do_error()



import argparse

import genv.sdk


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv detach" arguments to a parser.
    """

    parser.add_argument("--index", type=int, help="Device index to detach from")


def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv detach" logic.
    """

    if not genv.sdk.active():
        raise RuntimeError("Not running in an active environment")

    genv.sdk.detach(args.index)



import argparse
import pkg_resources


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """Adds "genv version" arguments to a parser."""

    pass


def run(args: argparse.Namespace) -> None:
    """Runs the "genv version" logic."""

    print(pkg_resources.get_distribution("genv").version)



import asyncio
from asyncio.subprocess import Process

from .runner import Runner as Base


class Runner(Base):
    async def _open_process(self, *args: str, stdin_fd: int, sudo: bool) -> Process:
        if sudo:
            args = ["sudo", *args]

        return await asyncio.create_subprocess_exec(
            *args,
            env=self._process_env,
            stdin=stdin_fd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )



import json
from typing import Iterable

from genv.entities.enforce import Report

from genv.serialization import JSONEncoder

from ..utils import reprint, run, Config, Command


async def execute(config: Config, reports: Iterable[Report]) -> None:
    """
    Executes reports on multiple hosts.

    :param reports: Reports to execute on each host

    :return: None
    """
    command = Command(["usage", "execute"], sudo=True)

    hosts, stdouts = await run(
        config,
        command,
        stdins=[json.dumps(report, cls=JSONEncoder) for report in reports],
    )

    reprint([host.hostname for host in hosts], stdouts)



import argparse
import os
import re
import socket
import shutil
from typing import Iterable, NoReturn, Optional

import psutil

from genv.entities import Env
import genv.sdk


def _find_port(env: Env) -> Optional[int]:
    """Finds any port an LLM server environment listens on."""

    match = re.match(r"^llm/[^/]+/(\d+)$", env.config.name)
    if match:
        return int(match.group(1))

    # this is a fallback for environments that were ran before 1.4.1
    for pid in env.pids:
        try:
            ports = genv.utils.get_process_listen_ports(pid)

            if len(ports) > 0:
                return ports[0]
        except psutil.NoSuchProcess:
            continue


def _find_available_port() -> int:
    """Finds an available port to listen on."""

    with socket.socket() as sock:
        sock.bind(("", 0))
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return sock.getsockname()[1]


def _exec_ollama(args: Iterable[str], host: str, port: int) -> NoReturn:
    """
    Executes ollama.
    Raises RuntimeError if ollama is not properly installed.

    :param args: Arguments to pass to ollama.
    :param host: Hostname to pass to ollama.
    :param port: Port to pass to ollama.
    """

    path = shutil.which("ollama")

    if not path:
        raise RuntimeError(
            """\
Could not find ollama.
You should install it if it's not already installed.
Otherwise, $PATH is probably not configured properly."""
        )

    os.execve(path, [path, *args], {"OLLAMA_HOST": f"{host}:{port}", **os.environ})


def do_attach(model: str) -> NoReturn:
    """
    Attaches to a running LLM.
    Raises RuntimeError if could not find the LLM.

    :param model: Model name.
    """

    with genv.utils.global_lock():
        envs = genv.core.envs.snapshot()

    for env in envs:
        if not env.config.name:
            continue

        if not (
            env.config.name.startswith(f"llm/{model}/")
            or env.config.name == f"llm/{model}"  # before 1.4.1
        ):
            continue

        port = _find_port(env)

        if port:
            # TODO(raz): we currently attach to the first replica; we should pick a replica in a smarter way.
            _exec_ollama(["run", model], host="localhost", port=port)

    raise RuntimeError(f"Could not find LLM model '{model}'")


def do_ps(format: str, header: bool, timestamp: bool) -> None:
    """
    Print LLM data in a human-readable fashion to the stdio.

    :param format: Print format ('csv' or 'tui').
    :param header: Print header row.
    :param timestamp: Print creation time as a timestamp instead of in a human-readable format.

    :return: None
    """
    with genv.utils.global_lock():
        envs = genv.core.envs.snapshot()

    if header:
        if format == "csv":
            print("MODEL,PORT,CREATED,EID,USER,PID(S)")
        elif format == "tui":
            print(
                "MODEL       PORT    CREATED              EID     USER            PID(S)"
            )

    for env in envs:
        if not (env.config.name and env.config.name.startswith("llm/")):
            continue

        model = env.config.name.split("/")[1]
        port = _find_port(env) or "N/A"
        created = env.creation if timestamp else genv.utils.time_since(env.creation)
        eid = env.eid
        user = f"{env.username}({env.uid})" if env.username else env.uid
        pids = " ".join(str(pid) for pid in env.pids)

        if format == "csv":
            print(f"{model},{port},{created},{eid},{user}{pids}")
        elif format == "tui":
            print(f"{model:<12}{port:<8}{created:<21}{eid:<8}{user:<16}{pids}")


def do_serve(
    model: str,
    host: str,
    port: Optional[int],
    gpus: Optional[int],
    gpu_memory: Optional[str],
) -> NoReturn:
    """Runs an LLM server in a newly created environment."""

    if not port:
        port = _find_available_port()

    with genv.sdk.env.activate(
        config=Env.Config(name=f"llm/{model}/{port}", gpus=gpus, gpu_memory=gpu_memory)
    ):
        _exec_ollama(["serve"], host=host, port=port)


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv llm" arguments to a parser.
    """

    subparsers = parser.add_subparsers(dest="command")

    def attach(parser):
        parser.add_argument("model")

    def ps(parser):
        parser.add_argument(
            "--no-header",
            dest="header",
            action="store_false",
            help="Do not print column headers",
        )
        parser.add_argument(
            "--timestamp", action="store_true", help="Print a non-prettified timestamp"
        )
        parser.add_argument(
            "--format",
            choices=["csv", "tui"],
            help="Output format; CSV or TUI (Text-based user interface)",
            default="tui",
        )

    def serve(parser):
        parser.add_argument("model")

        configuration = parser.add_argument_group("configuration")
        configuration.add_argument(
            "--host", default="127.0.0.1", help="Network interface to bind"
        )
        configuration.add_argument("-p", "--port", type=int, help="Port to bind")

        env = parser.add_argument_group("env")
        env.add_argument("--gpus", type=int, help="Device count")
        env.add_argument("--gpu-memory", help="Device memory capacity (e.g. 4g)")

    for command, help in [
        (attach, "Attach to a running LLM model"),
        (ps, "Print information about active LLMs"),
        (serve, "Start an LLM server"),
    ]:
        command(subparsers.add_parser(command.__name__, help=help))


def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv llm" logic.
    """

    if args.command == "attach":
        do_attach(args.model)
    elif args.command == "ps":
        do_ps(args.format, args.header, args.timestamp)
    elif args.command == "serve":
        do_serve(args.model, args.host, args.port, args.gpus, args.gpu_memory)
    else:
        do_ps(format="tui", header=True, timestamp=False)



import inspect
import sys

from genv.utils import memory_to_bytes

from .spec import Spec, Type


def General(*args, **kwargs) -> Spec:
    """Returns a general metric specification."""

    return Spec(Type.General, *args, **kwargs)


INSTALLATION = General("genv_is_installed", "Genv installation status")


def Device(*args, **kwargs) -> Spec:
    """
    Returns a per-device metric specification.
    """
    labelnames = kwargs.pop("labelnames", tuple())

    return Spec(
        Type.Device,
        *args,
        **kwargs,
        labelnames=("index",) + labelnames,
    )


DEVICE_TEMPERATURE = Device(
    "genv_device_temperature",
    "Device temperature in degrees C",
    convert=lambda device: device.temperature,
)

DEVICE_UTILIZATION = Device(
    "genv_device_utilization",
    "Device utilization",
    convert=lambda device: device.utilization,
)

DEVICE_MEMORY_USED = Device(
    "genv_device_memory_used_bytes",
    "Device used memory in bytes",
    convert=lambda device: memory_to_bytes(device.used_memory),
)

DEVICE_MEMORY_TOTAL = Device(
    "genv_device_memory_total_bytes",
    "Device total memory in bytes",
    convert=lambda device: memory_to_bytes(device.total_memory),
)


def System(*args, **kwargs) -> Spec:
    """
    Returns a system-wide metric specification.
    """
    return Spec(Type.System, *args, **kwargs)


ENVIRONMENTS = System(
    "genv_environments_total",
    "Number of active environments",
    convert=lambda snapshot: len(snapshot.envs),
)

PROCESSES = System(
    "genv_processes_total",
    "Number of running processes",
    convert=lambda snapshot: len(snapshot.processes),
)

ATTACHED_DEVICES = System(
    "genv_attached_devices_total",
    "Number of attached devices",
    convert=lambda snapshot: len(snapshot.devices.filter(attached=True)),
)

USERS = System(
    "genv_users_total",
    "Number of active users",
    convert=lambda snapshot: len(snapshot.envs.usernames),
)


def Environment(*args, **kwargs) -> Spec:
    """
    Returns a per-environment metric specification.
    """
    labelnames = kwargs.pop("labelnames", tuple())

    return Spec(
        Type.Environment,
        *args,
        **kwargs,
        labelnames=("eid",) + labelnames,
        filter=lambda labels, snapshot: labels[0] in snapshot.envs.eids,
    )


ENVIRONMENT_PROCESSES = Environment(
    "genv_environment_processes_total",
    "Number of running processes in an environment",
    convert=lambda snapshot: len(snapshot.processes),
)

ENVIRONMENT_ATTACHED_DEVICES = Environment(
    "genv_environment_attached_devices_total",
    "Number of attached devices of an environment",
    convert=lambda snapshot: len(snapshot.devices),
)


def Process(*args, **kwargs) -> Spec:
    """
    Returns a per-process metric specification.
    """
    labelnames = kwargs.pop("labelnames", tuple())

    return Spec(
        Type.Process,
        *args,
        **kwargs,
        labelnames=("pid", "eid") + labelnames,
        filter=lambda labels, snapshot: int(labels[0]) in snapshot.processes.pids,
    )


PROCESS_DEVICES = Process(
    "genv_process_devices_total",
    "Number of devices used by a process",
)

PROCESS_USED_GPU_MEMORY = Process(
    "genv_process_used_gpu_memory_bytes",
    "Used GPU memory by a process",
    labelnames=("device",),
)


def User(*args, **kwargs) -> Spec:
    """
    Returns a per-user metric specification.
    """
    labelnames = kwargs.pop("labelnames", tuple())

    return Spec(
        Type.User,
        *args,
        **kwargs,
        labelnames=("username",) + labelnames,
        filter=lambda labels, snapshot: labels[0] in snapshot.envs.usernames,
    )


USER_ENVIRONMENTS = User(
    "genv_user_environments_total",
    "Number of active environments of a user",
    convert=lambda snapshot: len(snapshot.envs),
)

USER_PROCESSES = User(
    "genv_user_processes_total",
    "Number of running processes of a user",
    convert=lambda snapshot: len(snapshot.processes),
)

USER_ATTACHED_DEVICES = User(
    "genv_user_attached_devices_total",
    "Number of attached devices of a user",
    convert=lambda snapshot: len(snapshot.devices),
)


ALL = [
    spec
    for _, spec in inspect.getmembers(
        sys.modules[__name__], lambda o: isinstance(o, Spec)
    )
]



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="genv",
    version="1.4.1",
    author="Run.ai",
    author_email="pypi@run.ai",
    description="GPU environment and cluster management",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/run-ai/genv",
    packages=setuptools.find_packages(),
    package_data={"genv": ["shims/*", "metrics/export/**/*"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: GPU",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: System :: Shells",
        "Topic :: Terminals",
    ],
    entry_points={
        "console_scripts": [
            "genv = genv.cli.__main__:main",
        ]
    },
    python_requires=">=3.7",
    install_requires="psutil",
    extras_require={
        "admin": ["prometheus_client"],
        "dev": ["black"],
        "monitor": ["prometheus_client"],
        "ray": ["ray", "pynvml"],
    },
)



import argparse

import genv.sdk


def add_arguments(parser: argparse.ArgumentParser) -> None:
    """
    Adds "genv home" arguments to a parser.
    """

    parser.add_argument("-q", "--quiet", action="store_true")


def run(args: argparse.Namespace) -> None:
    """
    Runs the "genv home" logic.
    """

    home = genv.sdk.home()

    if home:
        print(home)
    elif not args.quiet:
        raise RuntimeError(
            """\
Could not find any Genv configuration directory.

Create a directory named '.genv' under one of the following directories:
  1) Current working directory (i.e. $PWD)
  2) Home directory (i.e. $HOME)
"""
        )



from typing import Iterable

import prometheus_client

from .spec import Spec


class Metric(prometheus_client.Gauge):
    """
    A Genv metric.
    """

    def __init__(
        self,
        name: str,
        documentation: str,
        labelnames: Iterable[str],
        *,
        spec: Spec,
        **kwargs,
    ):
        super().__init__(name, documentation, labelnames, **kwargs)
        self._kwargs["spec"] = self.spec = spec

    @property
    def name(self) -> str:
        return self._name

    def labels(self, *labelvalues, **labelkwargs):
        """
        Sets metric label values if passed.
        """
        if labelvalues or labelkwargs:
            return super().labels(*labelvalues, **labelkwargs)

        return self

    def label_sets(self) -> Iterable[str]:
        """Returns all metric labels."""

        if not self._is_parent():
            return []

        with self._lock:
            return list(self._metrics.keys())



import asyncio
from abc import ABC, abstractmethod
from asyncio.subprocess import Process
from typing import Dict, Optional


class CommandResults:
    process: Process
    stdout: str
    stderr: str

    def __init__(self, process: Process, stdout: str, stderr: str):
        self.process = process
        self.stdout = stdout
        self.stderr = stderr


class Runner(ABC):
    _process_env: Dict[str, str]

    def __init__(self, process_env: Optional[Dict] = None):
        self._process_env = process_env

    async def run(
        self,
        *args: str,
        stdin: Optional[str] = None,
        sudo: bool = False,
        check: bool = False,
    ) -> CommandResults:
        stdin_fd = asyncio.subprocess.PIPE if stdin else asyncio.subprocess.DEVNULL
        process = await self._open_process(*args, stdin_fd=stdin_fd, sudo=sudo)

        stdout, stderr = await process.communicate(
            stdin.encode("utf-8") if stdin else None
        )

        stdout = stdout.decode("utf-8").strip()
        stderr = stderr.decode("utf-8").strip()

        if check and process.returncode != 0:
            command = " ".join(args)
            stdin_str = " with stdin " + str(stdin)
            raise RuntimeError(
                f"Failed running '{command}' {' with sudo ' if sudo else ''} { stdin_str if stdin else '' } ({stderr})"
            )

        return CommandResults(process, stdout, stderr)

    @abstractmethod
    async def _open_process(self, *args: str, stdin_fd: int, sudo: bool) -> Process:
        raise NotImplementedError("This should be implemented in subclasses")



from dataclasses import dataclass
import sys
from typing import Iterable, Optional

import genv.utils


@dataclass
class Process:
    """
    A compute running process.
    """

    @dataclass
    class Usage:
        """
        The GPU index and amount of GPU memory used by the process.
        """

        index: int
        gpu_memory: str

        @property
        def bytes(self) -> int:
            return genv.utils.memory_to_bytes(self.gpu_memory)

    pid: int
    used_gpu_memory: Iterable[Usage]
    eid: Optional[str]

    @property
    def indices(self) -> Iterable[int]:
        return [usage.index for usage in self.used_gpu_memory]

    @property
    def total_bytes(self) -> int:
        return sum(usage.bytes for usage in self.used_gpu_memory)

    def __hash__(self) -> int:
        return self.pid.__hash__()

    def filter(self, *, index: int):
        """
        Returns a new process with information only of the given device index.
        """
        return Process(
            self.pid,
            [usage for usage in self.used_gpu_memory if usage.index == index],
            self.eid,
        )

    @staticmethod
    def eid(pid: int) -> Optional[str]:
        """
        Returns the environment identifier of the process with the given pid.
        Returns None if the process is not running in an environment or if it could not be queried.
        """
        try:
            return genv.utils.get_process_environ(pid).get("GENV_ENVIRONMENT_ID")
        except PermissionError:
            print(
                f"[WARNING] Not enough permissions to query environment of process {pid}",
                file=sys.stderr,
            )
        except FileNotFoundError:
            print(f"[DEBUG] Process {pid} already terminated", file=sys.stderr)


@dataclass
class Processes:
    """
    A collection of processes.
    """

    processes: Iterable[Process]

    @property
    def pids(self) -> Iterable[int]:
        return [process.pid for process in self.processes]

    def __iter__(self):
        return self.processes.__iter__()

    def __len__(self):
        return self.processes.__len__()

    def __getitem__(self, pid: int) -> Process:
        return next(process for process in self.processes if process.pid == pid)

    def filter(
        self,
        deep: bool = True,
        *,
        pids: Optional[Iterable[int]] = None,
        eid: Optional[str] = None,
        eids: Optional[Iterable[str]] = None,
        index: Optional[int] = None,
    ):
        """
        Returns a new filtered collection.

        :param deep: Perform deep filtering
        :param pids: Process identifiers to keep
        :param eid: Environment identifier to keep
        :param eids: Environment identifiers to keep
        :param username: Username to keep
        """
        if eids:
            eids = set(eids)

        if eid:
            if not eids:
                eids = set()

            eids.add(eid)

        processes = self.processes

        if pids is not None:
            processes = [process for process in processes if process.pid in pids]

        if eids is not None:
            processes = [process for process in processes if process.eid in eids]

        if index is not None:
            if deep:
                processes = [process.filter(index=index) for process in processes]

            processes = [process for process in processes if index in process.indices]

        return Processes(processes)



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar, Union

import genv.utils

T = TypeVar("T")


class File(Generic[T], ABC):
    """
    A state file on disk.
    """

    def __init__(self, path: str, cleanup: bool = True, reset: bool = False) -> None:
        self._path = path
        self._cleanup = cleanup
        self._reset = reset

    @abstractmethod
    def _create(self) -> T:
        """Creates a new state."""
        pass

    @abstractmethod
    def _clean(self, state: T) -> None:
        """Cleans up state."""
        pass

    def _convert(self, o: Union[Any, T]) -> T:
        """Converts the loaded object if needed."""
        return o

    def load(self) -> T:
        """Loads state from disk."""
        self._state = genv.utils.load_state(
            self._path,
            creator=self._create,
            cleaner=self._clean,
            converter=self._convert,
            json_decoder=genv.serialization.JSONDecoder,
            cleanup=self._cleanup,
            reset=self._reset,
        )

        return self._state

    def save(self) -> None:
        """Saves state to disk."""
        genv.utils.save_state(
            self._state,
            self._path,
            json_encoder=genv.serialization.JSONEncoder,
        )

    def __enter__(self) -> T:
        return self.load()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()



from genv.entities.enforce import Survey


def env_devices(*surveys: Survey) -> None:
    """
    Terminates processes on devices not attached to their environments.
    """
    for survey in surveys:
        for env in survey.snapshot.envs:
            allowed = survey.snapshot.devices.filter(eid=env.eid).indices

            for process in survey.snapshot.processes.filter(eid=env.eid):
                unallowed = [index for index in process.indices if index not in allowed]

                if len(unallowed) == 0:
                    continue

                print(
                    f"{f'[{survey.hostname}] ' if survey.hostname else ''}"
                    f"Process {process.pid} from environment {env.eid} is using non-attached GPU(s) {','.join([str(index) for index in unallowed])}"
                )

                survey.terminate(process.pid)
